/* Padico Advanced Examples
 * author: Christian Prez
 */

#include <stdio.h>
#include <mpi.h>

#include <vector>

#include "XServiceIDL.h"
#include "XServiceType.h"

#include "DistributionBloc.h"

//#undef NO_COM

/************************************************************/
/************************************************************/
/************************************************************/

typedef struct schedule_s {
  
  int from;
  int to;

  int start;
  int len;

  int darrayid;
} schedule_t;

typedef vector<schedule_t> schedules_t;

class ScheduleHelper {

  unsigned int sprocid;
  unsigned int stotal;
  unsigned int sbsz;

  unsigned int dtotal;

public:
  ScheduleHelper(unsigned int id, unsigned int st, unsigned int sb, unsigned int dt)
    : sprocid(id), stotal(st), sbsz(sb), dtotal(dt) {}

  void compute_receive(unsigned int darrayid, schedules_t& sched_r, unsigned int low, unsigned int high, unsigned int base) {

    // compute process id of the sender
    unsigned int fpid = low / sbsz;
    unsigned int lpid = (high-1) / sbsz;

    // for each source pid, compute what we have to receive
    for(unsigned int i=fpid; i <= lpid; i++) {

      if ( i != sprocid ) {

	sched_r.resize(sched_r.size()+1);
	schedule_t& s =  sched_r[sched_r.size()-1];

	s.darrayid = darrayid;

	s.from  = i;
	unsigned int tmp = i*sbsz;
	s.start = (low >= tmp)?low:tmp; // max

	tmp = (i+1)*sbsz;
	unsigned int end = (high <= tmp)?high:tmp; // min
	s.len   = end - s.start;
	
	s.start -= base;
      }
    }
  }
};


/************************************************************/
/************************************************************/
/************************************************************/

extern "C" {

  void computeSendDataBlock1DSchedule(xservice_data_t *sdata, unsigned int glen, unsigned int total, unsigned int rank,
				      unsigned int dtotal, vector<PaCo::DistributionBlock1DLong>& vdarray, vector<int>& destid, 
				      schedules_t& sched_send, schedules_t& sched_recv) {
    
    // perform a local redistribution before sending data to server
    // if there is more server nodes than source node, data a 
    // distributed with a cyclic distribution on source side.
    
    // source data: sdata, bsz, sllen, glen, total, rank
    unsigned int bsz = blockSize(glen, total);
    
    unsigned long sl, sh;
    
    computeBlockBoundsO(&sl, &sh, glen, rank, bsz);
    unsigned sllen = localBlockLengthO(glen, rank, total, bsz);
    
    fprintf(stderr,"local data [ %ld, %ld [ llen = %d sbsz = %d\n", sl, sh, sllen, bsz);

    // remote data: vdarray, dl, gh, dbsz, dllen
    
    ////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////
    // compute how many local darray
    unsigned int nvdarray  = dtotal/total;
    unsigned int remaining = (nvdarray*total + rank < dtotal)?1:0;
    nvdarray += remaining;
    
    fprintf(stderr,"[%2d/%2d] nvdarray %d (dest total %d)\n\n", rank, total, nvdarray, dtotal);
    
    unsigned int dbsz = (glen + dtotal - 1) / dtotal; 
    
    // allocate memory
    vdarray.resize(nvdarray);
    destid.resize(nvdarray);
    
    // compute bounds
    for(unsigned int i=0; i< nvdarray; i++) {
      
      int procid = total*i+rank;

      destid[i] = procid;

      vdarray[i].topo.procid = procid;
      vdarray[i].topo.nbproc = dtotal;
      
      vdarray[i].dist.glow   = 0;
      vdarray[i].dist.ghigh  = glen;
      vdarray[i].dist.glen   = glen;
      
      computeBlockBoundsO(&vdarray[i].dist.llow, &vdarray[i].dist.lhigh, glen, procid, dbsz);
      
      vdarray[i].dist.llen   = vdarray[i].dist.lhigh - vdarray[i].dist.llow;
      
      fprintf(stderr,"remote data [ %ld, %ld [ llen = %ld\n", vdarray[i].dist.llow, vdarray[i].dist.lhigh, vdarray[i].dist.llen);
    }

    ////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////
    // locally redistribute data
  
    if (total == dtotal) { 
      // easy case: same bloc size -> Id redistribution 
      sched_send.reserve(0);
      sched_recv.reserve(0);
      for(unsigned int i=0; i< nvdarray; i++) {
	vdarray[i].data.replace(sllen, sllen, sdata);
      }      
    } else {
      fprintf(stderr,"\n--- XService::mySendDataBlock1D different bloc size...\n");
            
      sched_send.reserve(dtotal);
      sched_recv.reserve(total);
      
      // Object to handle receive schedule
      ScheduleHelper srecv(rank, total, bsz, dtotal);
      
      ////////////////////////////////////////////////////////////
      // filling what to send
      unsigned int fpid = sl / dbsz;
      unsigned int lpid = (sh-1) / dbsz;
      
      // for each dest bloc
      for(unsigned int i=fpid; i <= lpid; i++) {
	
	if ( (i % total) != rank) {

	  sched_send.resize(sched_send.size()+1);
	  schedule_t& s =  sched_send[sched_send.size()-1];
	  
	  s.to    = i % total;
	  unsigned int tmp = i*dbsz;	
	  s.start = ( sl >= tmp)?sl:tmp; // max
	  
	  tmp = (i+1)*dbsz;
	  unsigned int end = ( sh <= tmp)?sh:tmp; // min
	  
	  s.len   = end - s.start;	
	  s.start -= sl;
	}
      }
    
      ////////////////////////////////////////////////////////////
      // filling what to receive
      for(unsigned int i=0; i< nvdarray; i++) {
	// vdarray[i].llow, lhigh, llen
	unsigned int low  = vdarray[i].dist.llow;
	unsigned int high = vdarray[i].dist.lhigh;
	
	unsigned int did = (i*total+rank);
	unsigned int dllen = localBlockLengthO(glen, did, dtotal, dbsz);
	
	vdarray[i].data.length(dllen);
	xservice_data_t *ddata  = vdarray[i].data.get_buffer();
	// just for testing:	for(unsigned int j=0;j<vdarray[i].data.length(); j++) ddata[j]=1000+j;
	
	// test whether we have to copy something
	if ((low<sh) && (high>sl)) {
	  int s = (low >= sl)?low:sl; // max
	  int l = ((high<=sh)?high:sh) - s; // min
	  fprintf(stderr,"filling %d -> %d : %d -> %d from %ld -> %ld\n", s, s+l, s-low, s-low+l, s-sl, s-sl+l);
	  memcpy(&ddata[s-low], &sdata[s-sl], l*sizeof(xservice_data_t));
	
	  // test whether we have to receive something
	  if (low < sl) {
	    // case 1: on the left
	    srecv.compute_receive(i, sched_recv, low, sl, low);
	  } else {
	    // case 2: on the right
	    srecv.compute_receive(i, sched_recv, sh, high, low);
	  }
	} else {
	  // case 3  : all
	  srecv.compute_receive(i, sched_recv, low, high, low);
	}
      }
    }
  }

  void doSchedule(xservice_data_t *sdata, vector<PaCo::DistributionBlock1DLong>& vdarray, schedules_t& sched_send, schedules_t& sched_recv) {
    
    if (sched_send.size() || sched_recv.size()) {
      ////////////////////////////////////////////////////////////
      // Sending data
      
#ifndef NO_COM
      MPI_Request sreq[sched_send.size()];
      MPI_Request rreq[sched_recv.size()];
      
      MPI_Status sstat[sched_send.size()];
      MPI_Status rstat[sched_recv.size()];
#endif
      
      // Post Asynchronous MPI receive
      for(unsigned int i=0; i < sched_recv.size(); i++) {
	int j = sched_recv[i].darrayid;
	xservice_data_t* ddata = vdarray[j].data.get_buffer();
	
	fprintf(stderr, "recv: schedr no=%d start=%d len=%d from=%d\n", i, sched_recv[i].start, sched_recv[i].len,  sched_recv[i].from);
	
#ifndef NO_COM
	int err = MPI_Irecv(&ddata[sched_recv[i].start], sched_recv[i].len, XSERVICE_MPI_T, sched_recv[i].from,
			    51, MPI_COMM_WORLD, &rreq[i]);
	if (err!= MPI_SUCCESS) {
	  cerr << "EROR IN MPI_Irecv: return value is "<<err<<endl;
	}
#endif
      }
      
      // Send data via MPI
      for(unsigned int i=0; i < sched_send.size(); i++) {
	
	fprintf(stderr, "send: scheds no=%d start=%d len=%d to=%d\n", i, sched_send[i].start, sched_send[i].len,  sched_send[i].to);
	
#ifndef NO_COM
	int err = MPI_Isend(&sdata[sched_send[i].start], sched_send[i].len, XSERVICE_MPI_T, sched_send[i].to,
			    51, MPI_COMM_WORLD, &sreq[i]);
	
	if (err!= MPI_SUCCESS) {
	  cerr << "EROR IN MPI_Isend: return value is "<<err<<endl;
	}
#endif	
      }
      
      // Wait all receive & send
#ifndef NO_COM
      int err;
      err = MPI_Waitall(sched_send.size(), sreq, sstat);
      if (err!= MPI_SUCCESS) {
	cerr << "EROR IN MPI_WaitAll for send: return value is "<<err<<endl;
      }
      err = MPI_Waitall(sched_recv.size(), rreq, rstat);
      if (err!= MPI_SUCCESS) {
	cerr << "EROR IN MPI_WaitAll for recv: return value is "<<err<<endl;
      }
#endif
      fprintf(stderr,"\n--- XService::mySendDataBlock1D different bloc size...done\n");
    }    
  }

  void computeSendDataBlock1DClient(xservice_data_t *sdata, unsigned int glen, unsigned int total, unsigned int rank,
				    unsigned int dtotal, vector<PaCo::DistributionBlock1DLong>& vdarray, vector<int>& destid) {


    // allocating schedules
    schedules_t sched_send;
    schedules_t sched_recv;
    
    computeSendDataBlock1DSchedule(sdata,  glen, total, rank, dtotal, vdarray, destid, sched_send, sched_recv);
    
    doSchedule(sdata, vdarray, sched_send, sched_recv);


  }

  ////////////////////////////////////////////////////////////
  ////////////////////////////////////////////////////////////
  ////////////////////////////////////////////////////////////

  void computeSendDataBlock1DServer(xservice_data_t *sdata, unsigned int glen, unsigned int total, unsigned int rank,
				    unsigned int dtotal, vector<PaCo::DistributionBlock1DLong>& vdarray, vector<int>& destid) {

    // send data of node_source[i] to server_node[i % #server_node]
    // then the server must perform a redistribution

    
    unsigned int nvdarray  = dtotal/total;
    unsigned int remaining = (nvdarray*total + rank < dtotal)?1:0;
    nvdarray += remaining;
        
    // source data: sdata, bsz, sllen, glen, total, rank
    unsigned int bsz = blockSize(glen, total);
    
    unsigned long sl, sh;
    
    computeBlockBoundsO(&sl, &sh, glen, rank, bsz);   
    unsigned sllen = localBlockLengthO(glen, rank, total, bsz);
    
    fprintf(stderr,"local data [ %ld, %ld [ llen = %d sbsz = %d\n", sl, sh, sllen, bsz);
    
    ////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////
    
    fprintf(stderr,"[%2d/%2d] nvdarray %d (dest total %d)\n\n", rank, total, nvdarray, dtotal);
        
    // allocate memory
    vdarray.resize(nvdarray);
    
    // compute bounds
    for(unsigned int i=0; i< nvdarray; i++) {
      
      int procid = total*i+rank;
      
      destid[i] = procid;

      vdarray[i].topo.procid = rank;
      vdarray[i].topo.nbproc = total;
      
      vdarray[i].dist.glow   = 0;
      vdarray[i].dist.ghigh  = glen;
      vdarray[i].dist.glen   = glen;

      if (i==0) {

	vdarray[i].dist.llow   = sl;
	vdarray[i].dist.lhigh  = sh;
	vdarray[i].dist.llen   = sllen;
	
	vdarray[i].data.replace(sllen, sllen, sdata);
      } else {
	vdarray[i].dist.llow   = 0;
	vdarray[i].dist.lhigh  = 0;
	vdarray[i].dist.llen   = 0;
      }

      fprintf(stderr,"remote data [ %ld, %ld [ llen = %ld\n", vdarray[i].dist.llow, vdarray[i].dist.lhigh, vdarray[i].dist.llen);
    }    
  }


  ////////////////////////////////////////////////////////////
  ////////////////////////////////////////////////////////////

  void computeSendDataBlock1DComm(xservice_data_t *sdata, unsigned int glen, unsigned int total, unsigned int rank,
				  unsigned int dtotal, vector<PaCo::DistributionBlock1DLong>& vdarray, vector<int>& destid) {
  }

  void computeSendDataBlock1D(xservice_data_t *sdata,  unsigned int glen, unsigned int total, unsigned int rank,
			      unsigned int dtotal, vector<PaCo::DistributionBlock1DLong>& vdarray, vector<int> destid,
			      PaCo::distLoc_t mode) {

    switch(mode) {
    case PaCo::ClientSide: computeSendDataBlock1DClient(sdata, glen, total, rank, dtotal, vdarray, destid); break;
    case PaCo::ServerSide: computeSendDataBlock1DServer(sdata, glen, total, rank, dtotal, vdarray, destid); break;
    case PaCo::CommSide:   computeSendDataBlock1DComm  (sdata, glen, total, rank, dtotal, vdarray, destid); break;	
    }
  }

}


/************************************************************/
/************************************************************/
/************************************************************/

#ifdef NO_COM
extern "C" {

  int simSendDataBlock1D(unsigned int glen, int stotal, int dtotal) {

    int bsz = blockSize(glen,stotal);

    for(int rank=0; rank< stotal; rank++) {

      cerr << endl << "------------------------------- " << rank << " / " << stotal << endl;

      xservice_data_t sdata[bsz];  // tant pis pour la place perdu ! ca devrait etre sllen
      vector<PaCo::DistributionBlock1DLong> vdarray;
      vector<int> destid;

      computeSendDataBlock1D(sdata, glen, stotal, rank, dtotal, vdarray, destid);

      ////////////////////////////////////////////////////////////
      ////////////////////////////////////////////////////////////
      // send data to remote nodes
      for(unsigned int i=0; i< vdarray.size(); i++) {
	cout << endl << "Dumping vdarray["<<i<<"]:\n";
	cout << "   procid/total: " <<  vdarray[i].topo.procid << " / " << vdarray[i].toponbproc << endl;
	cout << " g low/high/len: " <<  vdarray[i].dist.glow << " / " << vdarray[i].dist.ghigh << " / " << vdarray[i].dist.glen << endl;
	cout << " l low/high/len: " <<  vdarray[i].dist.llow << " / " << vdarray[i].dist.lhigh << " / " << vdarray[i].dist.llen << endl;
	
	//      xserviceList[destidy[i]]->sendData(vdarray[i]);
      }
    }
    
    return 0;
  }
}

/************************************************************/

int main(int argc, char** argv) {

  if (argc != 4) {
    fprintf(stderr,"Usage: %s len source_total dest_total_nodes\n", argv[0]);
    exit(1);
  }
  
  unsigned int glen, stotal, dtotal;
  
  sscanf(argv[1],"%d",&glen);
  sscanf(argv[2],"%d",&stotal);
  sscanf(argv[3],"%d",&dtotal);

  fprintf(stderr,"Distribution block 1D with len=%d stotal=%d dtotal=%d\n", glen, stotal, dtotal);

  simSendDataBlock1D(glen, stotal, dtotal);
}
#endif
