/* Padico Advanced Examples
 * author: Christian Prez
 */

#include <stdio.h>
#include <mpi.h>

#include "Schedule.h"
#include "DistributionBloc.h"

#include "XServiceType.h"

//#define NO_COM

#define STANDALONE_FILE

#undef DEBUG_INTERNA
#undef DEBUG_COMM

#ifdef NO_COM
inline char* mymalloc(unsigned sz) {
  static unsigned base=0x1000000;
  unsigned ret = base;
  base+=0x1000;
  return (char*) ret;
}
#endif

#ifdef STANDALONE_FILE
int pid;

#include "Concret.h"

#endif

/************************************************************/
/************************************************************/
/************************************************************/

inline static unsigned getProcRangeInf(unsigned low, unsigned bsz) {
  return low / bsz;      // first remote node
}
inline static unsigned getProcRangeSup(unsigned high, unsigned bsz) {
  return (high-1) / bsz; //  last remote node
}

inline static unsigned getProcId(unsigned rank, Topology_t topo) {
  return (rank % topo.total);
}

/************************************************************/
/************************************************************/
/************************************************************/

// vOut represents what localData/stopo have to send to nodes of dtopo (vOut[].rank is in dtopo space)
void computeSendBlock1D(const GlobalData_t& gd, const LocalData_t &sd, 
			const Topology_t &stopo, const Topology_t &dtopo,
			vector<LocalData_t>& vOut) {
  
#ifdef DEBUG_INTERNAL
  cerr << "\nIn compute Send Schedule--------------------\n";
#endif

  if (stopo.total == dtopo.total) {
    //    vOut.reserve(vOut.size()+dtopo.total); // in bloc mode, at most one msg to each dest node
    vOut.push_back(sd);
#ifdef DEBUG_INTERNAL
    fprintf(stderr, "%d: rank:%d start:%d len:%d base:%p\n", pid, sd.rank, sd.start, sd.len, sd.base);
#endif
  } else {
    // Append mode
    vOut.reserve(vOut.size()+dtopo.total); // in bloc mode, at most one msg to each dest node

    unsigned dbsz = blockSize(gd.len, dtopo.total);
    
    unsigned long slow  = sd.start;
    unsigned long shigh = slow + sd.len;
    
    unsigned fpid, lpid;
    fpid = getProcRangeInf(slow,  dbsz);
    lpid = getProcRangeSup(shigh, dbsz);
    
    // for each dest bloc
    for(unsigned i=fpid; i <= lpid; i++) {           

      vOut.resize(vOut.size()+1);
      LocalData_t& s =  vOut[vOut.size()-1];
	
      s.rank    = i;
      unsigned tmp = i*dbsz;	
      s.start = ( slow >= tmp)?slow:tmp; // max
      
      tmp = (i+1)*dbsz;
      unsigned end = ( shigh <= tmp)?shigh:tmp; // min
      
      s.len   = end - s.start;

      s.base = sd.base + ((s.start - sd.start) * gd.unit_size);

#ifdef DEBUG_INTERNAL
      fprintf(stderr, "%d: s: to:%d start:%d len:%d base:%p\n", pid, s.rank, s.start, s.len, s.base);
#endif
    }
  }
}

////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////

// vOut represents what localData/dtopo have to receive from nodes of stopo (vOut[].rank is in stopo space)
void computeReceiveBlock1D(const GlobalData_t& gd, const LocalData_t &dd, 
			   const Topology_t &stopo, const Topology_t &dtopo,
			   vector<LocalData_t>& vOut) {

#ifdef DEBUG_INTERNAL
  cerr << "\nIn compute Receive Schedule--------------------\n";
#endif

  if (stopo.total == dtopo.total) {
    vOut.push_back(dd);
#ifdef DEBUG_INTERNAL
    fprintf(stderr, "%d: rank:%d start:%d len:%d base:%p\n", pid, dd.rank, dd.start, dd.len, dd.base);
#endif
  } else {
    // Apend mode
    vOut.reserve(vOut.size()+stopo.total); // in bloc mode, at most one msg from each src node

    unsigned sbsz = blockSize(gd.len, stopo.total);
    
    unsigned long dlow  = dd.start;
    unsigned long dhigh = dlow + dd.len;
    
    unsigned fpid, lpid;
    fpid = getProcRangeInf(dlow,  sbsz);
    lpid = getProcRangeSup(dhigh, sbsz);
    
    // for each dest bloc
    for(unsigned i=fpid; i <= lpid; i++) {
      
      vOut.resize(vOut.size()+1);
      LocalData_t& s =  vOut[vOut.size()-1];
	
      s.rank       = i;
      unsigned tmp = i*sbsz;	
      s.start = ( dlow >= tmp)?dlow:tmp; // max
	
      tmp = (i+1)*sbsz;
      unsigned end = ( dhigh <= tmp)?dhigh:tmp; // min
      
      s.len  = end - s.start;

      s.base = dd.base + ((s.start - dd.start) * gd.unit_size);

#ifdef DEBUG_INTERNAL
      fprintf(stderr, "%d: r: from:%d start:%d len:%d base:%p\n", pid, s.rank, s.start, s.len, s.base);
#endif
    }
  }
}

/************************************************************/
/************************************************************/
/************************************************************/

// ctopo is the current topo
// sdata: a pointer to the source data (pointer to local data)
// ddata: a pointer to the base address of of reveive (pointer to local data)
void doSchedule(const GlobalData_t& gd, const LocalData_t& ld, const Topology_t &ctopo,
		vector<LocalData_t>& sched_send, vector<LocalData_t>& sched_recv) {
  
  cerr << "\nIn doSchedule--------------------\n";

  if (sched_send.size() || sched_recv.size()) {

#ifndef NO_COM
    MPI_Request sreq[sched_send.size()];
    MPI_Request rreq[sched_recv.size()];
    unsigned si, ri;
    si=0;
    ri=0;
    
    MPI_Status sstat[sched_send.size()];
    MPI_Status rstat[sched_recv.size()];
#endif

    vector<LocalData_t*> local_recv;
    vector<LocalData_t*> local_send;

    ////////////////////////////////////////////////////////////
    // Sending data
        
    // Post Asynchronous MPI receive
#ifdef DEBUG_COMM
    cerr << pid << ": #sched_recv: " << sched_recv.size() << endl;
#endif
    for(unsigned i=0; i < sched_recv.size(); i++) {
      unsigned from = getProcId(sched_recv[i].rank, ctopo);
      if (from == ld.rank) {
#ifdef DEBUG_COMM
	fprintf(stderr, "%d: recv: schedr no=%d start=%d len=%d from=%d LOCAL\n", pid, i,
		sched_recv[i].start, sched_recv[i].len, from);
#endif
	local_recv.push_back(&sched_recv[i]);
      } else {
#ifdef DEBUG_COMM
	fprintf(stderr, "%d, recv: schedr no=%d start=%d len=%d from=%d base=%p\n", pid, i,
		sched_recv[i].start, sched_recv[i].len, from, sched_recv[i].base);
#endif
      
#ifndef NO_COM
	int err = MPI_Irecv(sched_recv[i].base, sched_recv[i].len*gd.unit_size, 
			    MPI_BYTE, from, 51, MPI_COMM_WORLD, &rreq[ri++]);
	if (err!= MPI_SUCCESS) {
	  cerr << "EROR IN MPI_Irecv: return value is "<<err<<endl;
	}
#endif
      }
    }
    
    // Send data via MPI
#ifdef DEBUG_COMM
    cerr << pid << ": #sched_send: " << sched_send.size() << endl;
#endif
    for(unsigned i=0; i < sched_send.size(); i++) {
      unsigned to = getProcId(sched_send[i].rank, ctopo);
      if (to == ld.rank) {
#ifdef DEBUG_COMM
	fprintf(stderr, "%d: send: scheds no=%d start=%d len=%d to=%d LOCAL\n", pid, i,
		sched_send[i].start, sched_send[i].len, to);
#endif
	local_send.push_back(&sched_send[i]);
      } else {
#ifdef DEBUG_COMM
	fprintf(stderr, "%d: send: scheds no=%d start=%d len=%d to=%d base=%p\n", pid, i, 
		sched_send[i].start, sched_send[i].len, to, sched_send[i].base);
#endif
      
#ifndef NO_COM
	int err = MPI_Isend(sched_send[i].base, sched_send[i].len*gd.unit_size, 
			    MPI_BYTE, to, 51, MPI_COMM_WORLD, &sreq[si++]);      
	if (err!= MPI_SUCCESS) {
	  cerr << "EROR IN MPI_Isend: return value is "<<err<<endl;
	}
#endif
      }
    }
    
    // Do local communication vie memcpy
    if (local_recv.size() != local_send.size()) {
      cerr << "Error: local recv & send have different size: " << local_recv.size() << " " << local_send.size() << endl;
    }
    for(unsigned i=0; i < local_recv.size(); i++) {
      if (local_recv[i]->len != local_send[i]->len) {
	cerr << "Error: local recv & send have different len for i= "<<i<< " :" << local_recv[i]->len << " " << local_send[i]->len << endl;
      }
#ifdef DEBUG_COMM
      fprintf(stderr, "%d: local: scheds no=%d start=%d len=%d\n", pid, i,
	      sched_send[i].start, sched_send[i].len);
#endif
#ifndef NO_COM
      memcpy(local_recv[i]->base, local_send[i]->base, local_send[i]->len*gd.unit_size);
#endif
    }


    // Wait all receive & send
#ifndef NO_COM
#ifdef DEBUG_INTERNAL
    cerr << "WAITING communication to end...\n";
#endif

    int err;
    err = MPI_Waitall(si, sreq, sstat);
    if (err!= MPI_SUCCESS) {
      cerr << "EROR IN MPI_WaitAll for send: return value is "<<err<<endl;
      }
    err = MPI_Waitall(ri, rreq, rstat);
    if (err!= MPI_SUCCESS) {
      cerr << "EROR IN MPI_WaitAll for recv: return value is "<<err<<endl;
    }
#ifdef DEBUG_INTERNAL
    cerr << "WAITING communication to end...ok \n";
#endif
#endif
  }
}

////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////

void computeSendDataBlock1DClient(const GlobalData_t& gd, const LocalData_t &sd, 
				  const Topology_t &stopo, const Topology_t &dtopo,
				  vAbstrait& vdarray, vector<unsigned>& destid)
{

  vector<LocalData_t> sched_send;
  vector<LocalData_t> sched_recv;

  sched_send.clear();
  sched_recv.clear();

  ////////////////////////////////////////////////////////////
  ////////////////////////////////////////////////////////////
  ////////////////////////////////////////////////////////////
  // Compute what to send
  ////////////////////////////////////////////////////////////
  computeSendBlock1D(gd, sd, stopo, dtopo, sched_send);

  ////////////////////////////////////////////////////////////
  ////////////////////////////////////////////////////////////
  ////////////////////////////////////////////////////////////
  // Compute what to receive
  ////////////////////////////////////////////////////////////
  // compute how many dest local data this node has to handle
  unsigned int nvdarray  = dtopo.total/stopo.total;
  unsigned int remaining = (nvdarray*stopo.total + sd.rank < dtopo.total)?1:0;
  nvdarray += remaining;

  // Preparing also CORBA stuff
  vdarray.clear();
  vdarray.resize(nvdarray);
  destid.clear();
  destid.resize(nvdarray);

  // so, we simulate node from rank, rank+stopo.total ,... only if nvdarray>0
  unsigned dbsz = blockSize(gd.len, dtopo.total);
  //  cerr << "DBSZ: " << dbsz << endl;
  for (unsigned it=0; it<nvdarray; it++) {

    Abstrait* out = vdarray[it];

    out->topo().total   = dtopo.total;
    out->gd().len       = gd.len;
    out->gd().unit_size = gd.unit_size;
    
    out->dist().length(1);
    out->seqDataLength(1);

    PaCo::PacoLocalData_t& pld = out->dist()[0];
    LocalData_t dd;
    pld.rank  = dd.rank  = sd.rank + it * stopo.total;
    pld.start = dd.start = dd.rank * dbsz;
    pld.len   = dd.len   = localBlockLengthO(gd.len, dd.rank, dtopo.total, dbsz);

#ifdef DEBUG_INTERNAL
    fprintf(stderr, "%d: -- r:%d gl:%d ll:%d start:%d bz:%d dt:%d\n", pid,
	    dd.rank, gd.len, dd.len, dd.start, dbsz, dtopo.total);
#endif

    destid[it] = dd.rank;

    ////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////
    // Allocating memory
#ifdef NO_COM
    dd.base = (char*) mymalloc(pld.len * gd.unit_size);
    out->DataLength(0, pld.len);
#else
#ifdef DEBUG_INTERNAL
    cerr << "Allocating a sequence of length " << pld.len << endl;
#endif
    out->DataLength(0, pld.len);
    if (pld.len > 0 )
      dd.base = out->DataGetBuffer(0, false);
    else
      dd.base = (char*) 1;
#endif

    if (dd.base==0) {
      cerr << "Cannot allocate memory for #elements" << pld.len  << endl;
    } else {
#ifdef DEBUG_INTERNAL
      fprintf(stderr, "dd.base = %p\n", dd.base);
#endif
    }

    computeReceiveBlock1D(gd, dd, stopo, dtopo, sched_recv);
  }

  ////////////////////////////////////////////////////////////
  ////////////////////////////////////////////////////////////
  ////////////////////////////////////////////////////////////
  // Doing local communication
  ////////////////////////////////////////////////////////////
  doSchedule(gd, sd, stopo, sched_send, sched_recv); 
}

////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////

// simply send data from source i to dest i % dtopo.total
// if less sources than destination: dest j <  source j % stopo.total

// Client Side function
void computeSendDataBlock1DServer(const GlobalData_t& gd, const LocalData_t &sd, 
				  const Topology_t &stopo, const Topology_t &dtopo,
				  vAbstrait& vdarray, vector<unsigned>& destid)
{

  vector<LocalData_t> sched_send;
  vector<LocalData_t> sched_recv;

  sched_send.clear();
  sched_recv.clear();

  ////////////////////////////////////////////////////////////
  ////////////////////////////////////////////////////////////
  ////////////////////////////////////////////////////////////
  // Compute what to send
  ////////////////////////////////////////////////////////////
  computeSendBlock1D(gd, sd, stopo, dtopo, sched_send);


  // How messages to send ?
  unsigned int nvdarray  = dtopo.total/stopo.total;
  unsigned int remaining = (nvdarray*stopo.total + sd.rank < dtopo.total)?1:0;
  nvdarray += remaining;

  // Preparing  CORBA stuff
  vdarray.clear();
  vdarray.resize(nvdarray);
  destid.clear();
  destid.resize(nvdarray);

  // so, we simulate node from rank, rank+stopo.total ,... only if nvdarray>0
  //  cerr << "DBSZ: " << dbsz << endl;
  for (unsigned it=0; it<nvdarray; it++) {

    Abstrait* out = vdarray[it];

    out->topo().total   = dtopo.total;
    out->gd().len       = gd.len;
    out->gd().unit_size = gd.unit_size;
    
    if (it >= 0 ) { // empty !
      out->dist().length(0);
      out->seqDataLength(0);
    } else { // it == 0
      out->dist().length(1);
      out->seqDataLength(1);

      out->dist()[0].rank  = sd.rank;
      out->dist()[0].start = sd.start;
      out->dist()[0].len   = sd.len;
      out->DataReplace(0, sd.len, sd.base, FALSE);
    }

    destid[it] = sd.rank + it * stopo.total;
  }
}

////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////

void computeSendDataBlock1DComm() {
}

////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////

// vOut represents what localData/stopo have to send to nodes of dtopo (vOut[].rank is in dtopo space)
void computeSendDataBlock1D(const GlobalData_t& gd, const LocalData_t &sd, 
			    const Topology_t &stopo, const Topology_t &dtopo,
			    vAbstrait& vdarray, vector<unsigned>& destid,
			    const PaCo::distLoc_t& mode) {

  switch(mode) {
  case PaCo::ClientSide: computeSendDataBlock1DClient(gd, sd, stopo, dtopo, vdarray, destid); break;
  case PaCo::ServerSide: computeSendDataBlock1DServer(gd, sd, stopo, dtopo, vdarray, destid); break;
  case PaCo::CommSide:   computeSendDataBlock1DComm  (); break;	
  }
  
}


////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

#ifdef STANDALONE_FILE
int simSendDataBlock1D(unsigned int glen, int total, int rank, int dtotal, const PaCo::distLoc_t& mode) {

  GlobalData_t gd;
  LocalData_t  sd;
  Topology_t   stopo, dtopo;
  
  vConcret vdarray;
  vector<unsigned> destid;
  
  gd.len       = glen;
  gd.unit_size = sizeof(xservice_data_t);
  
  unsigned bsz = blockSize(glen, total);
    
  sd.rank  = rank;
  sd.start = computeBlockBoundInf0(bsz, rank);
  sd.len   = localBlockLengthO(glen, rank, total, bsz);
#ifdef NO_COM
  sd.base  = (char*) 0x1000;
#else
  sd.base  = (char*) malloc(sd.len*gd.unit_size);
  xservice_data_t* p=(xservice_data_t*) sd.base;
  for( unsigned i=0; i < sd.len; i++) {
    *p++ = sd.start+i;
  }
#endif
  
#if 0
  cerr << "Dumping data: ";
  for(unsigned k=0; k < sd.len; k++)
    cerr << " " << ((xservice_data_t*)sd.base)[k];
  cerr << endl;
#endif

  stopo.total = total;
  dtopo.total = dtotal;
  
  computeSendDataBlock1D(gd, sd, stopo, dtopo, vdarray, destid, mode);
  
  ////////////////////////////////////////////////////////////
  ////////////////////////////////////////////////////////////
  // send data to remote nodes

  MPI_Barrier(MPI_COMM_WORLD);

  cerr << "\n #vdarray: " << vdarray.size() << "\n";
  for(unsigned i=0; i< vdarray.size(); i++) {
    cout << "Dumping vdarray["<<i<<"] to " << destid[i] << " :\n";
    cout << "   topo / gd : " <<  vdarray[i]->topo().total << " / " << vdarray[i]->gd().len << " / " << vdarray[i]->gd().unit_size << endl;
    cout << " dist # " <<  vdarray[i]->dist().length() << endl;
    for(unsigned j=0; j< vdarray[i]->dist().length(); j++) {
      cout << " rank/low/len (len): " <<  vdarray[i]->dist()[j].rank << " / " << vdarray[i]->dist()[j].start << " / " << vdarray[i]->dist()[j].len << " ( " << vdarray[i]->getDataLength(j) << " )" <<endl;
    }
      
#ifndef NO_COM
    //        cerr << " data: "; fprintf(stderr, "%p:", &vdarray[i].Data(j, 0);
    //        for(unsigned k=0; k < vdarray[i].getDataLength(j); k++)
    //  	cerr << " " << vdarray[i].Data(j, k);
#endif
    cerr << endl;
    
    //      pglobal->xserviceList[destid[i]]->sendData(vdarray[i]);
    
  }   
  return 0;
}

int main(int argc, char** argv) {

  MPI_Init(&argc, &argv);

  if (argc != 4) {
    fprintf(stderr,"Usage: %s len source_total dest_total_nodes\n", argv[0]);
    exit(1);
  }
  
  unsigned glen, stotal, dtotal;
  
  sscanf(argv[1],"%d",&glen);
  sscanf(argv[2],"%d",&stotal);
  sscanf(argv[3],"%d",&dtotal);

  fprintf(stderr,"Distribution block 1D with len=%d stotal=%d dtotal=%d\n", glen, stotal, dtotal);

 int rank;
 MPI_Comm_rank(MPI_COMM_WORLD, &rank);
 pid=rank;

  //  for(unsigned rank=0; rank< stotal; rank++) {
 cerr << endl << "------------------------------- " << rank << " / " << stotal << endl;
 simSendDataBlock1D(glen, stotal, rank, dtotal, PaCo::ClientSide);
    //  }

 cerr << pid << ": ENDING !!\n";

  MPI_Finalize();
}
#endif
