/* Padico Advanced Examples
 * author: Christian Prez
 */

#include <Padico/Puk.h>
#include <Padico/Module.h>
#include <Padico/CORBA.h>

#include <stdio.h>
#include <mpi.h>

#include <vector>

#include "XServiceCorba.h"
#include "XService.h"

#include "DistributionBloc.h"
#include "Schedule.h"

#include "Concret.h"

/************************************************************/
/************************************************************/
/************************************************************/

/* PRIAVTE !!! */

typedef struct myglobal_s {
  
  // For every client/server
  int counter;
  
  Padico::ORB_ptr                orb;
  CosNaming::NamingContext_var   naming_context;
  
  PaCo::distLoc_t                distMode;
  
  // Only for client
  Server1D_ptr                   xserviceProxy;
  vector<Server1D_ptr>           xserviceList;
  int rank;
  
  // Only for server
    PortableServer::POA_var        poa;
  PortableServer::POAManager_var poa_manager;
  CosNaming::Name                server_name;
  PortableServer::ObjectId_var   server_oid;
  
  XServiceIDL_impl*              server;
} myglobal_t;

myglobal_t* pglobal;

/* END PRIVATE !!! */


/************************************************************/
/************************************************************/
/************************************************************/

PADICO_MODULE_DECLARE(XService);
PADICO_MODULE_INIT(XService_internal_init);
PADICO_MODULE_FINALIZE(XService_internal_finalize);

/************************************************************/

int XService_internal_init(void)
{
  padico_print("XService: module init...\n");

  pglobal = new myglobal_t();
  pglobal->server=0;
  pglobal->counter=0;

  pglobal->orb = Padico::ORB_init();

  try {
    CORBA::Object_var ns_ref = pglobal->orb->resolve_initial_references ("NameService");
    pglobal->naming_context = CosNaming::NamingContext::_narrow (ns_ref);
  } 
  catch(...) {
    padico_print("XService: cannot resolve NameService.\n");
    return 101;
  }

  padico_print("XService: module init...done\n");

  return 0;
}

/************************************************************/

void XService_internal_finalize(void)
{
  padico_print("XService: module finalize (counter= %d)\n",pglobal->counter);
}


/************************************************************/
/************************************************************/
/************************************************************/


extern "C" {

  void sendRedistributionInfo() {
    int rank;
    MPI_Barrier(MPI_COMM_WORLD); 
    
    MPI_Comm_rank(MPI_COMM_WORLD,&rank);

    if (rank==0) {
      for(unsigned int i=0; i< pglobal->xserviceList.size(); i++) {
	pglobal->xserviceList[i]->setLocationRedistribution(pglobal->distMode);
      }
    }

    MPI_Barrier(MPI_COMM_WORLD);
  }

  void setRedistributionClient() { pglobal->distMode = PaCo::ClientSide; sendRedistributionInfo();}
  void setRedistributionServer() { pglobal->distMode = PaCo::ServerSide; sendRedistributionInfo();}
  void setRedistributionComm()   { pglobal->distMode = PaCo::CommSide;   sendRedistributionInfo();}

}

/************************************************************/
/************************************************************/
/************************************************************/

extern "C" {
  int register1D(int rank, int total, char *name)
  {
    if (pglobal->server!=0) return 1;

    CORBA::Object_var poa_ref = pglobal->orb->resolve_initial_references ("RootPOA");
    pglobal->poa = PortableServer::POA::_narrow (poa_ref);
    pglobal->poa_manager = pglobal->poa->the_POAManager();
    pglobal->poa_manager->activate();

    // Creating the servant
    pglobal->server = new XServiceIDL_impl();

    // Activating the servant
    pglobal->server_oid = pglobal->poa->activate_object(pglobal->server);
  
    // Getting its IOR
    CORBA::Object_var ref = pglobal->poa->id_to_reference (pglobal->server_oid.in());
    CORBA::String_var str = pglobal->orb->object_to_string (ref.in());

    // Init interfaceParallel
    pglobal->server->init_InterfaceParallel(rank, total, str.in());

    if (rank==0) {
      pglobal->server_name.length (1);
      pglobal->server_name[0].id = CORBA::string_dup(name);
      pglobal->server_name[0].kind = CORBA::string_dup("");
      pglobal->naming_context->rebind(pglobal->server_name, pglobal->server->_this());

      cerr << "R) Dumping servername: " << pglobal->server_name[0].id << ":" << pglobal->server_name[0].id<< endl;
    }

    MPI_Barrier(MPI_COMM_WORLD);

    return 0;
  }

} /* extern "C" */

/************************************************************/
 
extern "C" {

  int unregister1D(int rank)
  {
    if (pglobal->server==0) return 1;

    if (rank==0) {
      try {
	cerr << "U) Dumping servername: " << pglobal->server_name[0].id << ":" << pglobal->server_name[0].id<< endl;
	pglobal->naming_context->unbind(pglobal->server_name);
      } catch (...) {
	cerr << "OOPS: catch exception when unbinding server_name\n";
      }
    }

    pglobal->poa->deactivate_object(pglobal->server_oid);

    delete pglobal->server;
    pglobal->server=0;

    return 0;
  }
  
} /* extern "C" */


/************************************************************/
/************************************************************/
/************************************************************/

extern "C" {

  int connect1D(char* name) {
    // Finding proxy
    CosNaming::Name                xservice_name;

    xservice_name.length (1);
    xservice_name[0].id = CORBA::string_dup(name);
    xservice_name[0].kind = CORBA::string_dup("");
    try {
      CORBA::Object_var server_obj = pglobal->naming_context->resolve(xservice_name);
      pglobal->xserviceProxy = Server1D::_narrow(server_obj);
    } 
    catch(...) {
      fprintf(stderr,"connect1D: cannot resolve %s in NameService.\n", name);
      exit(-1);
    }  

    // Getting infos
    cerr << "looking for connections infos\n";
    PaCo::ConnectionInfos_var ci;
    try {
      ci = pglobal->xserviceProxy->getConnectionInfos();
    } catch(...) {
      cerr << "ERROR: get an exception !\n";
      exit(-1);
    }
    cerr << "looking for connections infos...ok\n";
    
    // Reserving memory
    pglobal->xserviceList.resize(ci->nodes.length());

    // Finding nodes
    for(unsigned int i=0; i<ci->nodes.length(); i++) {
      CORBA::Object_var _obj = pglobal->orb->string_to_object (ci->nodes[i]);
      pglobal->xserviceList[i] = Server1D::_narrow(_obj);
    }     

    pglobal->distMode=pglobal->xserviceProxy->getLocationRedistribution();

    // Waiting every body
    MPI_Barrier(MPI_COMM_WORLD);
    cerr << "Connect: ok\n";
    return 0;
  }
}

/************************************************************/
/************************************************************/
/************************************************************/

extern "C" {
  void mykillServer() {
    // synchro client side
    MPI_Barrier(MPI_COMM_WORLD); 

    for(unsigned int i=0; i< pglobal->xserviceList.size(); i++) {
      pglobal->xserviceList[i]->kill();
    }
  }
}

/************************************************************/
/************************************************************/
/************************************************************/

extern "C" {

  int mySendDataBlock1D(xservice_data_t *sdata, unsigned int glen, int rank, int total) {

    GlobalData_t gd;
    LocalData_t  sd;
    Topology_t   stopo, dtopo;

    vConcret vdarray;
    vector<unsigned> destid;

    gd.len       = glen;
    gd.unit_size = sizeof(xservice_data_t);

    unsigned bsz = blockSize(glen, total);

    sd.rank  = rank;
    sd.start = computeBlockBoundInf0(bsz, rank);
    sd.len   = localBlockLengthO(glen, rank, total, bsz);
    sd.base  = (char*) sdata;

    stopo.total = total;
    dtopo.total = pglobal->xserviceList.size(); 

    computeSendDataBlock1D(gd, sd, stopo, dtopo, vdarray, destid, pglobal->distMode);

    ////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////

    cerr << "Doing CORBA Call\n";
    // send data to remote nodes
    for(unsigned i=0; i< vdarray.size(); i++) {
      cout << "Dumping vdarray["<<i<<"] to " << destid[i] << " :\n";
      cout << "   topo / gd : " <<  vdarray[i]->topo().total << " / " << vdarray[i]->gd().len << " / " << vdarray[i]->gd().unit_size << endl;
      cout << " dist # " <<  vdarray[i]->dist().length() << endl;
      for(unsigned j=0; j< vdarray[i]->dist().length(); j++) {
	cout << " rank/low/len (len): " <<  vdarray[i]->dist()[j].rank << " / " << vdarray[i]->dist()[j].start << " / " << vdarray[i]->dist()[j].len << " ( " << vdarray[i]->getDataLength(j) << " )" <<endl;
      }
      
      Concret* c = (Concret*)vdarray[i];

      pglobal->xserviceList[destid[i]]->sendData( c->getData(0));
    }    
    return 0;
  }
}

/************************************************************/

extern "C" {

  int myWaitDataBlock1D(xservice_data_t **data, unsigned int* llen, unsigned int *glen) {

    return pglobal->server->waitData1DBlock(data, llen, glen);
    
  }
}

/************************************************************/
/************************************************************/
/************************************************************/

extern "C" {

  xservice_data_t checksum(xservice_data_t  *data, int len) {

    xservice_data_t cs=0;

    for(int i=0; i < len; i++) {
      cs+=data[i];
    }
    return cs;
  }
}
