      program main  
************************************************************ 
* program to solve a finite difference 
* discretization of Helmholtz equation :  
* (d2/dx2)u + (d2/dy2)u - alpha u = f 
* using Jacobi iterative method. 
*
* Modified: Sanjiv Shah,       Kuck and Associates, Inc. (KAI), 1998
* Author:   Joseph Robicheaux, Kuck and Associates, Inc. (KAI), 1998
* 
* Directives are used in this code to achieve paralleism. 
* All do loops are parallized with default 'static' scheduling.
* 
* Input :  n - grid dimension in x direction 
*          m - grid dimension in y direction
*          alpha - Helmholtz constant (always greater than 0.0)
*          tol   - error tolerance for iterative solver
*          relax - Successice over relaxation parameter
*          mits  - Maximum iterations for iterative solver
*
* On output 
*       : u(n,m) - Dependent variable (solutions)
*       : f(n,m) - Right hand side function 
*************************************************************
      implicit none        

      integer nmax,mmax
      parameter (nmax=2049)
      parameter (mmax=2049)

      integer n,m,mits,mtemp 
      double precision tol,relax,alpha 
      common /idat/ n,m,mits,mtemp
      common /fdat/tol,alpha,relax

      integer nproc, me
      common/processor/nproc, me
      integer blocksize, blockbegin, blockend
      common/distribution/blocksize, blockbegin, blockend
      integer momemyapplicationnode, momenbrapplicationnodes
      integer momenbrdsmnodes, momemydsmnode
      integer momebroadcastint
      double precision momebroadcastdouble

      integer testbroadcast(10)

      print *, "JacobiMome is starting..."

      call momesetalternatesignalstack(0)
      print *, "JacobiMome: 1"
      call MomeInit()
      print *, "JacobiMome: 2"
      call momegroupbarrier(0)

      print *, "JacobiMome: 3"

      me= momemyapplicationnode()
      nproc= momenbrapplicationnodes()

      print *, "JacobiMome: 4"

c      print *, ' node ', me, ' nproc ', nproc

c      testbroadcast(1)= me
c      testbroadcast(2)= me+1
c      testbroadcast(5)= -1
c      mits= 8
c      call momegroupbroadcastbuf(1, 0, testbroadcast(1),
c     - testbroadcast(5), 8, mits)
c      print *, ' node ', me, mits, testbroadcast(5), testbroadcast(6)
 

 
* 
* Read info 
*     
      if(me.eq.0)then         
      if(me.eq.1)then
         write(*,*) "Input n,m - grid dimension in x,y direction " 
         read(5,*) n,m 
      else
         n= 64
         m= 64
      end if
      if(n.gt.nmax) then
         print *, "max for n : ", nmax
         stop
      endif
      if(m.gt.mmax) then
         print *, "max for m : ", mmax
         stop
      endif
      if(me.eq.1)then
         write(*,*) "Input alpha - Helmholts constant " 
         read(5,*) alpha
         write(*,*) "Input relax - Successive over-relaxation parameter"
         read(5,*) relax 
         write(*,*) "Input tol - error tolerance for iterative solver"
         read(5,*) tol 
         write(*,*) "Input mits - Maximum iterations for solver" 
         read(5,*) mits
      else
         alpha= 1.0
         relax= 1.0
         tol= 0.00000000001
         mits= 1
      end if         
      endif
      
      call momegroupbarrier(0)

c      print *, me, ' n ? ', n
      n= momebroadcastint(0,n)
      call momegroupbarrier(0)
c      print *, me, ' n ', n
      m= momebroadcastint(0,m)
      call momegroupbarrier(0)
c      print *, me, ' m ', m
      alpha= momebroadcastdouble(0,alpha)
      relax= momebroadcastdouble(0,relax)
      tol= momebroadcastdouble(0,tol)
      mits= momebroadcastint(0,mits)

      blocksize= (n+nproc-1)/nproc
      blockbegin= me*blocksize + 1
      blockend= min(blockbegin+blocksize-1,n)

c      print *, me, ' block ', blockbegin, blockend
*
* Calls a driver routine 
* 
      call driver () 


c      print *, ' termin ', me
      call MomeFinalize ()
      stop
      end 

      subroutine driver ( ) 
*************************************************************
* Subroutine driver () 
* This is where the arrays are allocated and initialzed. 
*
* Working varaibles/arrays 
*     dx  - grid spacing in x direction 
*     dy  - grid spacing in y direction 
*************************************************************
      implicit none 

      integer nmax,mmax
      parameter (nmax=2049)
      parameter (mmax=2049)
      integer n,m,mits,mtemp 
      double precision tol,relax,alpha 
      common /idat/ n,m,mits,mtemp
      common /fdat/tol,alpha,relax


      integer momesize, numseg, ptalloc, allocated 

      integer nproc, me
      common/processor/nproc, me
      integer blocksize, blockbegin, blockend
      common/distribution/blocksize, blockbegin, blockend
      integer momeallocatesharedsegment, momememorymappingsize

      double precision wu(nmax,mmax),wf(nmax,mmax),uold(nmax,mmax)
      integer iaa, iab, iac, iad 
      common /arrays/iaa(2048),wu,
     /               iab(2048),wf,
     /               iac(2048),uold,
     /               iad(2048)

      double precision dx,dy

c      print *, me, ' momememorymappingsize... '
      ptalloc= 0

      momesize= momememorymappingsize(wu(1,1),uold(nmax,mmax))
      if (me.eq.0) then
         print *, me, ' size jacobi ', momesize
      endif
 
c      print *, me, ' momeallocatesharedsegment... '
      numseg= momeallocatesharedsegment(momesize)

      if (me.eq.0) then
         print *, me, ' segment jacobi ', numseg, ' size ', momesize
      endif

      call momemapmemorysection(wu(1,1),uold(nmax,mmax), numseg,
     /                  ptalloc, allocated)
      

      call momestrongconsistency(wu(1,1),uold(nmax,mmax))
c      call mometracethepages(wu(1,1),uold(nmax,mmax))
* Initialize data

      call initialize (n,m,alpha,dx,dy,wu,wf)

* Solve Helmholtz equation

      call jacobi (n,m,dx,dy,alpha,relax,wu,wf,tol,mits)

* Check error between exact solution

      call  error_check (n,m,alpha,dx,dy,wu,wf)

      return 
      end 

      subroutine initialize (n,m,alpha,dx,dy,u,f) 
******************************************************
* Initializes data 
* Assumes exact solution is u(x,y) = (1-x^2)*(1-y^2)
*
******************************************************
      implicit none 
     
      integer nmax,mmax
      parameter (nmax=2049)
      parameter (mmax=2049)

      integer nproc, me
      common/processor/nproc, me
      integer blocksize, blockbegin, blockend, bj, ej
      common/distribution/blocksize, blockbegin, blockend

      integer n,m
      double precision u(nmax,mmax),f(nmax,mmax)
      double precision wu(nmax,mmax),wf(nmax,mmax),uold(nmax,mmax)
      integer iaa, iab, iac, iad
      common /arrays/iaa(2048),wu,
     /               iab(2048),wf,
     /               iac(2048),uold,
     /               iad(2048)

      double precision dx,dy,alpha
      
      integer i,j, xx,yy
      double precision PI 
      parameter (PI=3.1415926)

      print *, me, ' init... '
      dx = 2.0 / (n-1)
      dy = 2.0 / (m-1)

* Initilize initial condition and RHS

!$omp parallel do private(xx,yy)
      call allocateiterations(1, m, bj, ej)
      do j = bj,ej
c         print *,me,j
         do i = 1,n
c            print *, me, ' iteration ', j, i
            xx = -1.0 + dx * dble(i-1)        ! -1 < x < 1
            yy = -1.0 + dy * dble(j-1)        ! -1 < y < 1
            u(i,j) = 0.0 
            f(i,j) = -alpha *(1.0-xx*xx)*(1.0-yy*yy) 
     &           - 2.0*(1.0-xx*xx)-2.0*(1.0-yy*yy)
         enddo
      enddo  
!$omp end parallel do
c      print *, me, ' init done '

      print *, me, ' init...done '
      return 
      end 

      subroutine jacobi (n,m,dx,dy,alpha,omega,u,f,tol,maxit)
******************************************************************
* Subroutine HelmholtzJ
* Solves poisson equation on rectangular grid assuming : 
* (1) Uniform discretization in each direction, and 
* (2) Dirichlect boundary conditions 
* 
* Jacobi method is used in this routine 
*
* Input : n,m   Number of grid points in the X/Y directions 
*         dx,dy Grid spacing in the X/Y directions 
*         alpha Helmholtz eqn. coefficient 
*         omega Relaxation factor 
*         f(n,m) Right hand side function 
*         u(n,m) Dependent variable/Solution
*         tol    Tolerance for iterative solver 
*         maxit  Maximum number of iterations 
*
* Output : u(n,m) - Solution 
*****************************************************************
      implicit none 
      integer nmax,mmax
      parameter (nmax=2049)
      parameter (mmax=2049)
      integer n,m,maxit
      double precision dx,dy, alpha, tol,omega

      double precision u(nmax,mmax),f(nmax,mmax)
      double precision wu(nmax,mmax),wf(nmax,mmax),uold(nmax,mmax)
      integer iaa, iab, iac, iad
      common /arrays/iaa(2048),wu,
     /               iab(2048),wf,
     /               iac(2048),uold,
     /               iad(2048)

      integer nproc, me
      common/processor/nproc, me
      integer blocksize, blockbegin, blockend, bj, ej
      common/distribution/blocksize, blockbegin, blockend
      double precision momesumdouble

      double precision starttime, stoptime, mometime, totaltime


*
* Local variables 
* 
      integer i,j,k,k_local 
      double precision error,resid,rsum,ax,ay,b
      double precision error_local

      integer momenbrdsmnodes, momemydsmnode
      real ta,tb,tc,td,te,ta1,ta2,tb1,tb2,tc1,tc2,td1,td2
      real te1,te2
      real second
      external second
*
* Initialize coefficients 
      totaltime= mometime()

      ax = 1.0/(dx*dx) ! X-direction coef 
      ay = 1.0/(dy*dy) ! Y-direction coef
      b  = -2.0/(dx*dx)-2.0/(dy*dy) - alpha ! Central coeff  


      error = 10.0 * tol 
      k = 1

      do while (k.le.maxit .and. error.gt. tol) 

         error = 0.0  

         starttime= mometime()

* Copy new solution into old
!$omp parallel

!$omp do 
         call allocateiterations(1, m, bj, ej)
         do j=bj,ej
            do i=1,n
               uold(i,j) = u(i,j) 
            enddo
         enddo

* Compute stencil, residual, & update

        call momesync()
!$omp do private(resid) reduction(+:error)
        call allocateiterations(2, m-1, bj, ej)
        call momesynchronize(uold(2,bj-1),uold(n-1,bj-1))
        call momesynchronize(uold(2,ej+1),uold(n-1,ej+1))
         do j = bj,ej
            do i = 2,n-1 
*     Evaluate residual 
               resid = (ax*(uold(i-1,j) + uold(i+1,j)) 
     &                + ay*(uold(i,j-1) + uold(i,j+1))
     &                 + b * uold(i,j) - f(i,j))/b
* Update solution 
               u(i,j) = uold(i,j) - omega * resid
* Accumulate residual error
               error = error + resid*resid 
            end do
         enddo
 

!$omp enddo nowait
!$omp end parallel

* Error check 
         error= momesumdouble(error) 

         k = k + 1
  

         error = sqrt(error)/dble(n*m)
         if(me.eq.0)then
            stoptime= mometime()-starttime
            print *, me, ' iteration ', k-1, 
     /                   ' error ', error, 
     /                   ' time ', stoptime 
         endif
*
      enddo                     ! End iteration loop 
*

      if(me.eq.0)then
         print *, 'Total time ', mometime()-totaltime
         print *, 'Total Number of Iterations ', k 
         print *, 'Residual                   ', error 
      endif
      return 
      end 

      subroutine error_check (n,m,alpha,dx,dy,u,f) 
      implicit none 
************************************************************
* Checks error between numerical and exact solution 
*
************************************************************ 

      integer nmax,mmax
      parameter (nmax=2049)
      parameter (mmax=2049)

      integer nproc, me
      common/processor/nproc, me
      integer blocksize, blockbegin, blockend, bj, ej
      common/distribution/blocksize, blockbegin, blockend

      double precision momesumdouble
     
      integer n,m
      double precision dx,dy,alpha 
      
      double precision u(nmax,mmax),f(nmax,mmax)
      double precision wu(nmax,mmax),wf(nmax,mmax),uold(nmax,mmax)
      integer iaa, iab, iac, iad
      common /arrays/iaa(2048),wu,
     /               iab(2048),wf,
     /               iac(2048),uold,
     /               iad(2048)

      integer i,j
      double precision xx,yy,temp,error 

      dx = 2.0 / (n-1)
      dy = 2.0 / (m-1)
      error = 0.0 

!$omp parallel do private(xx,yy,temp) reduction(+:error)
      call allocateiterations(1, m, bj, ej)
      do j = bj,ej
         do i = 1,n
            xx = -1.0d0 + dx * dble(i-1)
            yy = -1.0d0 + dy * dble(j-1)
            temp  = u(i,j) - (1.0-xx*xx)*(1.0-yy*yy)
            error = error + temp*temp 
         enddo
      enddo

      error= momesumdouble(error)
  
      error = sqrt(error)/dble(n*m)

      print *, 'Solution Error : ',error

      return 
      end 
 
      subroutine allocateiterations(from, to, first, last)

      integer nproc, me
      common/processor/nproc, me
      integer blocksize, blockbegin, blockend
      common/distribution/blocksize, blockbegin, blockend

      integer from, to, first, last


c      call momemutexlock (10)
      first= max(from, blockbegin)
      last=  min(to,   blockend)
c      call momemutexunlock (10)
      end
   
