/** @file
 * @brief Padico Circuit Service.
 * @ingroup Circuit
 */

/** @defgroup Circuit Service: Circuit -- SPMD-oriented, Madeleine-like network access
 * @author Alexandre Denis
 */

#include <Padico/Puk.h>
#include <Padico/Topology.h>
#include <Padico/Module.h>
#include "Circuit-internals.h"

PADICO_MODULE_DECLARE(Circuit);
PADICO_MODULE_INIT(circuit_init);

/* ********************************************************* */

static padico_circuit_list_t circuits;

static int circuit_init(void)
{
  puk_iface_register("Circuit");
  circuits = padico_circuit_list_new();
  padico_circuit_builder_init();
  return 0;
}

#define foreach_circuit_adapter(CIRCUIT, FUNC) \
{ puk_adapter_itor_t z; \
  puk_vect_foreach(z, puk_adapter, (CIRCUIT)->assembly_patterns) { \
    const struct padico_circuit_driver_s*driver = puk_component_get_driver_Circuit(*z, NULL); \
    if(driver->FUNC)(*driver->FUNC)(CIRCUIT); \
  } \
}

/* *** Circuit management *********************************** */

padico_circuit_t padico_circuit_init(const char*group_id, const char*circuit_id)
{
  int i;
  padico_circuit_t c = padico_circuit_lookup(circuit_id);
  padico_group_t   g = padico_group_lookup(group_id);
  if(g == NULL)
    {
      padico_warning("Circuit: can't create circuit -- group does not exist [ GroupID=%s ]\n",
		     group_id);
      return NULL;
    }
  if(c != NULL)
    {
      padico_warning("Circuit: trying to create duplicate circuit %s.\n", circuit_id);
      return c;
    }
  padico_out(50, "creating new circuit \"%s\" based on group %s\n", circuit_id, group_id);
  /* allocate a new circuit */
  c = padico_circuit_new();
  c->circuit_id = padico_strdup(circuit_id);
  c->group      = g;
  c->callback   = NULL;
  c->cb_key     = NULL;
  /* allocate empty out connections */
  c->assembly_patterns = puk_adapter_vect_new();
  puk_adapter_specific_init(&c->as);
  c->outs = padico_malloc(padico_circuit_size(c)*sizeof(struct padico_circuit_connection_s));
  for(i = 0; i < padico_circuit_size(c); i++)
    {
      padico_na_node_t node = padico_group_node(c->group, i);
      puk_adapter_t assembly = padico_ns_serial_selector(node, NULL, puk_iface_Circuit());
      if(!puk_adapter_vect_find(c->assembly_patterns, assembly))
	puk_adapter_vect_push_back(c->assembly_patterns, assembly);
      padico_circuit_connection_init(&c->outs[i], c, assembly, i);
    }
  /* init all sub-adapters */
  foreach_circuit_adapter(c, init);
  padico_circuit_list_push_back(circuits, c);
  /* build the outgoing connections */
  padico_circuit_build_connections(c);
  /* The circuit is ready! */
  padico_out(30, "circuit **%s** succesfully created!\n", c->circuit_id);
  padico_out(30, padico_string_get(padico_circuit_serialize(c)));
  return c;
}

void padico_circuit_start(padico_circuit_t c,
			  padico_circuit_cb_t callback,
			  void*key)
{
  padico_out(30, "entering circuit_id=**%s**\n", c->circuit_id);
  c->callback = callback;
  c->cb_key = key;
  foreach_circuit_adapter(c, start);
  padico_out(30, "exiting circuit_id=**%s**\n", c->circuit_id);
}

void padico_circuit_shutdown(padico_circuit_t c)
{
  padico_out(30, "entering circuit_id=**%s**\n", c->circuit_id);
  foreach_circuit_adapter(c, stop);
  padico_out(30, "exiting circuit_id=**%s**\n", c->circuit_id);
}

void padico_circuit_close(padico_circuit_t c)
{
  padico_out(30, "entering circuit_id=**%s**\n", c->circuit_id);
  foreach_circuit_adapter(c, stop);
  foreach_circuit_adapter(c, close);
  padico_circuit_list_erase(circuits, c);
  padico_circuit_delete(c);
  padico_out(30, "exiting circuit_id=**%s**\n", c->circuit_id);
}

padico_circuit_t padico_circuit_lookup(const char*id)
{
  padico_circuit_itor_t i;
  for(i  = padico_circuit_list_begin(circuits);
      i != padico_circuit_list_end(circuits);
      i  = padico_circuit_list_next(i))
    {
      if(strcmp(i->circuit_id, id) == 0)
	break;
    }
  return i;
}

padico_circuit_t padico_circuit_hook(const char*circuit_id, const char*module_id, int header)
{
  const padico_circuit_t circuit = padico_circuit_lookup(circuit_id);
  if(!circuit)
    {
      /** @todo instead of lookup+fatal error, we should check that our module
       * was loaded by the Padico SPMDizer proxy with the correct initialization
       * procedure, and issue an informative message if it is not the case.
       */
      padico_fatal("%s: circuit %s not found.\n", module_id, circuit_id);
    }
  padico_print("Circuit: hook module %s; hdr_size=%d\n", module_id, header);
  return circuit;
}

