/** @file
 * @ingroup NetSelector-basic
 */


#include <Padico/NetSelector.h>
#include "Ns-basic-internals.h"

PADICO_MODULE_HOOK(NetSelector_basic);


/* *** Debug *********************************************** */

#ifndef NDEBUG

static void ns_basic_print_entities(ns_entity_vect_t entity_vect)
{
  ns_entity_t*_n;
  puk_vect_foreach(_n, ns_entity, entity_vect)
    {
      ns_entity_t n = *_n;
      switch(n->kind)
	{
	case NS_ENTITY_NODE:
	  padico_print("NetSelector: node: %s\n", n->node.name);
	  break;
	case NS_ENTITY_HOST:
	  padico_print("NetSelector: host: %s\n", n->host.name);
	  break;
	case NS_ENTITY_GROUP:
	  padico_print("NetSelector: group: %s\n", n->group.name);
	  ns_basic_print_entities(&n->group.entities);
	  break;
	default:
	  padico_fatal("NetSelector-basic: bad entity type.\n");
	  break;
	}
    }
}

static void ns_basic_print_groups(puk_hashtable_t groups)
{
  puk_hashtable_enumerator_t e =
    puk_hashtable_enumerator_new(groups);
  char*group_name;
  ns_entity_t group;
  puk_hashtable_enumerator_next2(e, &group_name, &group);
  while(group)
    {
      padico_print("NetSelector: group: %s [root]\n", group->group.name);
      ns_basic_print_entities(&group->group.entities);
      puk_hashtable_enumerator_next2(e, &group_name, &group);
    }
}

static void ns_basic_rule_print_entity(ns_entity_t i)
{
  switch(i->kind)
    {
    case NS_ENTITY_NODE:
      padico_print("NetSelector:      -- node: %s\n", i->node.name?:"*");
      break;
    case NS_ENTITY_HOST:
      padico_print("NetSelector:      -- host: %s\n", i->host.name?:"*");
      break;
    case NS_ENTITY_GROUP:
      padico_print("NetSelector:      -- group: %s\n", i->group.name);
      break;
    default:
      padico_print("NetSelector:      -- [unknown kind]\n");
    }
}

static void ns_basic_print_rules(void)
{
  ns_rule_t rule;
  puk_hashtable_enumerator_t rb_enum =
    puk_hashtable_enumerator_new(nsdb.rulebuckets);
  struct ns_rule_bucket_s*bucket;
  ns_rule_list_t rules;
  puk_hashtable_enumerator_next2(rb_enum, (const void**)&bucket, (void**)&rules);
  while(bucket)
    {
      puk_list_foreach(rule, rules)
	{
	  padico_print("NetSelector: ***** rule- profile=%s; iface=%s\n",
		       bucket->profile, bucket->iface->name);
	  padico_print("NetSelector: * target:\n");
	  ns_target_itor_t target;
	  puk_vect_foreach(target, ns_target, &rule->targets)
	    {
	      switch(target->kind)
		{
		case NS_TARGET_CLIQUE:
		  padico_print("NetSelector:   -- clique:\n");
		  ns_basic_rule_print_entity(target->clique.entity);
		  break;
		case NS_TARGET_MULTIPARTITE:
		  {
		    ns_entity_t*_i;
		    padico_print("NetSelector:   -- multipartite:\n");
		    puk_vect_foreach(_i, ns_entity, target->multipartite.subsets)
		      {
			ns_entity_t i = *_i;
			ns_basic_rule_print_entity(i);
		      }
		  }
		  break;
		case NS_TARGET_UNKNOWN:
		  padico_print("NetSelector:   -- unknown\n");
		  break;
		default:
		  padico_fatal("NetSelector: bad target type %d in rule.\n", target->kind);
		  break;
		}
	    }
	  padico_print("NetSelector: * assembly:\n");
	  padico_string_t s = puk_adapter_serialize(rule->assembly);
	  padico_print("NetSelector: %s\n", padico_string_get(s));
	  padico_string_delete(s);
	  padico_print("NetSelector: ***** rule end\n");
	}
      puk_hashtable_enumerator_next2(rb_enum, (const void**)&bucket, (void**)&rules);
    }
}

void ns_basic_debug_init(void)
{
  const char*debug = padico_getattr("NS_BASIC_DEBUG");
  const int show_all = debug?(strstr(debug, "all") != NULL):0;
  if(debug)
    {
      padico_print("NetSelector: ***** debug NS_BASIC_DEBUG=%s\n", debug);
      if(show_all || strstr(debug, "nodes"))
	{
	  puk_hashtable_enumerator_t e =
	    puk_hashtable_enumerator_new(nsdb.nodes_by_name);
	  char*node_name;
	  ns_entity_t node;
	  padico_print("NetSelector: *** Listing nodes ***\n");
	  puk_hashtable_enumerator_next2(e, &node_name, &node);
	  while(node_name)
	    {
	      padico_print("NetSelector: node: %s (%s)\n", node_name, node->node.name);
	      puk_hashtable_enumerator_next2(e, &node_name, &node);
	    }
	}
      if(show_all || strstr(debug, "clusters"))
	{
	  padico_print("NetSelector: *** Listing groups ***\n");
	  ns_basic_print_groups(nsdb.groups_by_name);
	}
      if(show_all || strstr(debug, "rules"))
	{
	  padico_print("NetSelector: *** Listing rules ***\n");
	  ns_basic_print_rules();
	}
    }
}
       
#endif /* NDEBUG */

