/** @file
 * @brief NetSelector with lazy name-resolution and lazy component loading.
 * @ingroup NetSelector
 */


#include "NetSelector.h"

#include <Padico/Module.h>
#include <Padico/Puk.h>

PADICO_MODULE_DECLARE(NetSelector);
PADICO_MODULE_INIT(ns_netselector_init);

/* ********************************************************* */

const static char default_profile[] = "default";

typedef const struct padico_netselector_driver_s*padico_netselector_driver_t;
/** describes an entry in the selectors hashtable
 */
struct netselector_entry_s
{
  const struct padico_netselector_driver_s*selector; /**< the driver for the given selector */
  const char*name; /**< the name for this entry- 
		    * @note this field may be different from selector->name (because of aliases)
		    */
};
PUK_VECT_TYPE(netselector_entry, struct netselector_entry_s);

static struct
{
  puk_hashtable_t table; /**< table of know selectors [ key: profile; data: netselector_entry_vect_t ] */
  marcel_mutex_t lock;   /**< lock for accessing the hashtable */
} nsdb;

static netselector_entry_vect_t ns_selector_lookup(const char*profile);
static padico_netselector_driver_t ns_selector_resolve(struct netselector_entry_s*e);


/* ********************************************************* */

static int ns_netselector_init(void)
{
  puk_iface_register("NetSelector");
  nsdb.table = puk_hashtable_new_string();
  marcel_mutex_init(&nsdb.lock, NULL);
  return 0;
}

void padico_ns_selector_register(const struct padico_netselector_driver_s*selector)
{
  padico_out(6, "adding selector=%s for profile=%s\n", selector->name, selector->profile?:"*");
  puk_component_declare(selector->name,
			puk_component_provides("NetSelector", "selector", selector));
}

/* ********************************************************* */

puk_adapter_t padico_ns_serial_selector(padico_na_node_t peer, const char*_profile, 
					puk_iface_t iface)
{
  const char*profile = _profile?:default_profile;
  const netselector_entry_vect_t v = ns_selector_lookup(profile);
  puk_adapter_t assembly = NULL;
  netselector_entry_vect_itor_t i;
  padico_out(20, "remote peer node=%s- computing assembly for iface=%s; profile=%s.\n", 
	     peer?padico_na_node_getname(peer):"*", iface->name, profile);
  assert(peer != NULL);
  puk_vect_foreach(i, netselector_entry, v)
    {
      const padico_netselector_driver_t selector = ns_selector_resolve(i);
      if(selector)
	assembly = (*selector->serial)(padico_na_getlocalnode(), peer, profile, iface);
      if(assembly)
	break;
    }
  return assembly;
}

puk_adapter_t padico_ns_host_selector(padico_na_host_t host, const char*_profile, 
				      puk_iface_t iface, uint16_t port, int protocol)
{
  const char*profile = _profile?:default_profile;
  const netselector_entry_vect_t v = ns_selector_lookup(profile);
  puk_adapter_t assembly = NULL;
  netselector_entry_vect_itor_t i;
  padico_out(20, "remote peer host=%s- computing assembly for iface=%s; profile=%s.\n", 
	     host?padico_na_host_getname(host):"*", iface->name, profile);
  puk_vect_foreach(i, netselector_entry, v)
    {
      const padico_netselector_driver_t selector = ns_selector_resolve(i);
      if(selector)
	{
	  padico_out(20, "trying selector *%s*\n", selector->name);
	  assembly = (*selector->host)(host, profile, iface, port, protocol);
	  if(assembly)
	    {
	      padico_out(20, "got answer from selector *%s*\n", selector->name);
	      break;
	    }
	}
    }
  return assembly;
}


/* ********************************************************* */

/** resolve a selector entry (load driver from name).
 */
static padico_netselector_driver_t ns_selector_resolve(struct netselector_entry_s*e)
{
  if(e->name && !e->selector)
    {
      const puk_adapter_t adapter = puk_adapter_resolve(e->name);
      if(!adapter)
	{
	  padico_warning("NetSelector: will not use selector *%s*.\n", e->name);
	  e->name = NULL;
	  return NULL;
	}
      e->selector = puk_component_get_driver_NetSelector(adapter, NULL);
      assert(e->selector);
    }
  return e->selector;
}

/** returns the known selector entries (vect) for a given profile
 */
static netselector_entry_vect_t ns_selector_lookup(const char*profile)
{
  netselector_entry_vect_t v = NULL;
  marcel_mutex_lock(&nsdb.lock);
  v = puk_hashtable_lookup(nsdb.table, profile);
  if(!v)
    {
      padico_string_t s = padico_string_new();
      const char*selector_name = NULL;
      int i = 0;
      padico_out(6, "resolving selectors for profile *%s*\n", profile);
      v = netselector_entry_vect_new();
      puk_hashtable_insert(nsdb.table, profile, v);
      do
      {
	padico_string_printf(s, "PADICO_NETSELECTOR_%s_%d", profile, i);
	selector_name = padico_getattr(padico_string_get(s));
	if(selector_name)
	  {
	    padico_print("NetSelector: resolving- %s:%d... %s\n", profile, i, selector_name);
	    const struct netselector_entry_s e = 
	      {
		.name = selector_name,
		.selector = NULL /* unresolved by default- selectors are lazy-loaded */
	      };
	    netselector_entry_vect_push_back(v, e);
	  }
	i++;
      }
      while(selector_name);
      padico_string_delete(s);
    }
  if(netselector_entry_vect_empty(v))
    {
      padico_fatal("NetSelector: no selector for profile *%s*.\n", profile);
    }
  marcel_mutex_unlock(&nsdb.lock);
  return v;
}

