/** @file psp-minidriver.c
 *  @brief PSP interface for nmad minidriver
 */

#include <Padico/Module.h>
#include <Padico/Puk.h>
#include <Padico/Topology.h>
#include <Padico/PSP.h>
#include <Padico/NetSelector.h>
#include <Padico/AddrDB.h>
#include <Padico/PadicoControl.h>

PADICO_MODULE_DECLARE(PSP_minidriver);
PADICO_MODULE_INIT(padico_psp_minidriver_init);

#if defined(NMAD)

#include <nm_minidriver.h>

#ifdef PIOMAN
#include <pioman.h>
#endif /* PIOMAN */

/* ********************************************************* */
/* *** PadicoSimplePackets for minidriver */

static void*psp_minidriver_instanciate(puk_instance_t instance, puk_context_t context);
static void psp_minidriver_destroy(void*_status);
static void psp_minidriver_init(void*_status, uint32_t tag, const char*label, size_t*h_size,
				padico_simplepackets_handler_t handler, void*key);
static void psp_minidriver_connect(void*_status, padico_na_node_t node);
static padico_simplepackets_connection_t psp_minidriver_new_message(void*_status, 
								    padico_na_node_t node, void**_sbuf);
static void psp_minidriver_pack(void*_status, padico_simplepackets_connection_t _conn,
			  const char*bytes, size_t size);
static void psp_minidriver_end_message(void*_status, padico_simplepackets_connection_t _conn);

/** instanciation facet for PSP-minidriver
 * @ingroup PSP-minidriver
 */
const static struct puk_adapter_driver_s psp_minidriver_adapter_driver =
  {
    .instanciate = &psp_minidriver_instanciate,
    .destroy     = &psp_minidriver_destroy
  };

/** 'SimplePackets' facet for PSP-minidriver
* @ingroup PSP-minidriver
*/
const static struct padico_simplepackets_driver_s psp_minidriver_driver =
  {
    .init        = &psp_minidriver_init,
    .connect     = &psp_minidriver_connect,
    .new_message = &psp_minidriver_new_message,
    .pack        = &psp_minidriver_pack,
    .end_message = &psp_minidriver_end_message
  };

/* ********************************************************* */
/* *** Structures */


PUK_VECT_TYPE(psp_minidriver_conns, struct psp_minidriver_status_s*);

static struct padico_minidriver_s
{
  struct psp_minidriver_conns_vect_s conns;    /**< established connections */
  struct padico_psp_directory_s slots;         /**< PSP demultiplexing */
  int polling_thread_launched;                 /**< if polling thread was started */
  marcel_mutex_t lock;                         /**< local access to global info (conns & slots) */
#ifdef PIOMAN
  struct piom_ltask ltask;
#endif /* PIOMAN */
} padico_minidriver;

struct psp_minidriver_status_s
{
  struct
  {
    struct puk_receptacle_NewMad_minidriver_s r;
    puk_context_t context;                     /**< sub-contexts for driver */
    struct nm_minidriver_properties_s props;   /**< driver properties */
    const void*url;                            /**< driver remote url */
  } trk[2];
  padico_na_node_t remote_node;                /**< remote padico node */
  padico_psp_slot_t slot;                      /**< PSP slot */
  size_t h_size;                               /**< user's requeted header size */
  char*sbuf;                                   /**< pre allocated header buffer for sending */
  char*rbuf;                                   /**< header buffer for receive */
  struct iovec v;                              /**< iovec for rbuf  */
  int header_sent;                             /**< state of the header */
  size_t url_size;
  marcel_mutex_t mutex;                        /**< connection mutex */
  int threshold;
  volatile int pending;
  volatile int posted;
};


/* ********************************************************* */

static void*psp_minidriver_instanciate(puk_instance_t instance, puk_context_t context)
{
  struct psp_minidriver_status_s*status = padico_malloc(sizeof(struct psp_minidriver_status_s));
  puk_context_indirect_NewMad_minidriver(instance, "trk0", &status->trk[0].r);
  puk_context_indirect_NewMad_minidriver(instance, "trk1", &status->trk[1].r);
  puk_component_conn_t trk0 = puk_context_conn_lookup(context, NULL, "trk0");
  puk_component_conn_t trk1 = puk_context_conn_lookup(context, NULL, "trk1");
  status->slot = NULL;
  status->trk[0].context = trk0->context;
  status->trk[1].context = trk1->context;
  (*status->trk[0].r.driver->getprops)(-1, &status->trk[0].props);
  size_t url0_size = 0, url1_size = 0;
  (*status->trk[0].r.driver->init)(status->trk[0].context, &status->trk[0].url, &url0_size);
  (*status->trk[1].r.driver->init)(status->trk[1].context, &status->trk[1].url, &url1_size);
  assert(url0_size == url1_size);
  status->url_size = url0_size;
  padico_print("PSP-minidriver: instanciate- status = %p; url_size = %d\n", status, (int)status->url_size);
  status->threshold = 8*1024;
  status->pending = 0;
  status->posted = 0;
  return status;
}

static void psp_minidriver_destroy(void*_status)
{
  struct psp_minidriver_status_s*status = (struct psp_minidriver_status_s*)_status;
  if(status->slot)
    padico_psp_slot_remove(&padico_minidriver.slots, status->slot);
  padico_free(status);
}

/* ********************************************************* */

static inline void psp_minidriver_lock(struct psp_minidriver_status_s*status)
{
#ifdef PIOMAN
  piom_ltask_mask(&padico_minidriver.ltask);
#endif /* PIOMAN */
  marcel_mutex_lock(&status->mutex);
}

static inline void psp_minidriver_unlock(struct psp_minidriver_status_s*status)
{
  marcel_mutex_unlock(&status->mutex);
#ifdef PIOMAN
  piom_ltask_unmask(&padico_minidriver.ltask);
#endif /* PIOMAN */
}

static void psp_minidriver_pump(void*_token, void*bytes, size_t size)
{
  struct psp_minidriver_status_s*status = _token;
  struct iovec v = { .iov_base = bytes, .iov_len = size };
  const int trk = (size > status->threshold) ? 1 : 0;
  /* pump size message in bytes ptr */
  (*status->trk[trk].r.driver->recv_init)(status->trk[trk].r._status, &v, 1);
  int rc = -1;
  do
    {
      rc = (*status->trk[trk].r.driver->poll_one)(status->trk[trk].r._status);
    }
  while(rc != 0);

}

static void psp_minidriver_sendheader(struct psp_minidriver_status_s*status)
{
  struct iovec v;
  status->header_sent = 1;
  v.iov_base = status->sbuf;
  v.iov_len = status->slot->h_size;
  (*status->trk[0].r.driver->send_post)(status->trk[0].r._status, &v, 1);
  int rc = -1;
  do
    {
      rc = (*status->trk[0].r.driver->send_poll)(status->trk[0].r._status);
    }
  while(rc != 0);
}

static void psp_minidriver_refill(struct psp_minidriver_status_s*status)
{
  status->pending = 0;
  *(uint32_t*)status->rbuf = 0xDEADBEEF;
  status->v.iov_base = status->rbuf;
  status->v.iov_len = status->slot->h_size;
  (*status->trk[0].r.driver->recv_init)(status->trk[0].r._status, &status->v, 1);
  status->posted = 1;
}

static struct psp_minidriver_status_s*psp_minidriver_poll(void)
{
  int rc = -1;
  struct psp_minidriver_status_s*status = NULL;
  psp_minidriver_conns_vect_itor_t i;
  puk_vect_foreach(i, psp_minidriver_conns, &padico_minidriver.conns)
    {
      status = *i;
      if(status->posted && !status->pending)
	{
#ifndef PIOMAN
	  marcel_mutex_lock(&status->mutex);
#endif /* PIOMAN */
	  rc = (*status->trk[0].r.driver->poll_one)(status->trk[0].r._status);
#ifndef PIOMAN
	  marcel_mutex_unlock(&status->mutex);
#endif /* PIOMAN */
	  if(rc == 0)
	    {
	      status->pending = 1;
	      status->posted = 0;
	      return status;
	    }
	}
      else if(status->pending)
	{
	  return status;
	}
    }
  return NULL;
}

#ifdef PIOMAN
static int psp_minidriver_task(void*_status)
{
  struct psp_minidriver_status_s*status = psp_minidriver_poll();
  if(status != NULL)
    {
      *(struct psp_minidriver_status_s**)_status = status;
      piom_ltask_completed(&padico_minidriver.ltask);
    }
  return 0;
}
#endif /* PIOMAN */

static void*psp_minidriver_worker(void*dummy)
{
#ifdef PIOMAN
  volatile struct psp_minidriver_status_s*_status = NULL;
  piom_ltask_create(&padico_minidriver.ltask, &psp_minidriver_task, &_status, PIOM_LTASK_OPTION_REPEAT, piom_topo_full);
#endif /* PIOMAN */
  for(;;)
    {
#ifdef PIOMAN
      piom_ltask_submit(&padico_minidriver.ltask);
      piom_ltask_wait(&padico_minidriver.ltask);
      struct psp_minidriver_status_s*status = (void*)_status;
#else /* PIOMAN */
      struct psp_minidriver_status_s*status = psp_minidriver_poll();
#endif /* PIOMAN */
      if(status)
	{
	  marcel_mutex_lock(&status->mutex);
	  const uint32_t*p_tag = (void*)status->rbuf;
	  assert(*p_tag == status->slot->tag);
	  const padico_psp_slot_t slot = padico_psp_slot_lookup(&padico_minidriver.slots, *p_tag);
	  (*(slot->handler))(p_tag + 1, status->remote_node, slot->key, &psp_minidriver_pump, status);
	  psp_minidriver_refill(status);
	  marcel_mutex_unlock(&status->mutex);
	}
      else
	{
	  sched_yield();
	}
    }
  return NULL;
}

/* ********************************************************* */

#endif /* NMAD */

extern int padico_psp_minidriver_init(void)
{
#if defined (NMAD)
  marcel_mutex_init(&padico_minidriver.lock, NULL);
  puk_component_declare("PSP-minidriver",
			puk_component_provides("PadicoAdapter", "adapter", &psp_minidriver_adapter_driver),
			puk_component_provides("PadicoSimplePackets", "psp", &psp_minidriver_driver),
			puk_component_uses("NewMad_minidriver", "trk0"),
			puk_component_uses("NewMad_minidriver", "trk1"));
  
  /* Initialize static structure */
  padico_psp_directory_init(&padico_minidriver.slots);
  psp_minidriver_conns_vect_init(&padico_minidriver.conns);
  padico_minidriver.polling_thread_launched = 0;
  return 0;
  
#else  /* NMAD */

  padico_warning("PSP-minidriver: not initialized- NewMad not included in core.");
  return -1;

#endif /* NMAD */
}

#if defined (NMAD)

/* ********************************************************* */

static void psp_minidriver_init(void*_status, 
				uint32_t tag, 
				const char*label, 
				size_t*h_size,
				padico_simplepackets_handler_t handler, 
				void*key)
{
  struct psp_minidriver_status_s*status = (struct psp_minidriver_status_s*)_status;
  /* Now we know what is the size of the buffer, we can allocate (and 1 uint32_t is kept for us) */
  *h_size += 16;
  status->h_size = *h_size;
  status->slot = padico_psp_slot_insert(&padico_minidriver.slots, handler, key, tag, status->h_size);
  status->sbuf = padico_malloc(*h_size);
  status->rbuf = padico_malloc(*h_size);
}

static void psp_minidriver_connect(void*_status, padico_na_node_t node)
{
  struct psp_minidriver_status_s*status = _status;
  status->remote_node = node;
  status->header_sent = 0;
  marcel_mutex_init(&status->mutex, NULL);
  padico_print("PSP-minidriver: connecting...\n");
  int trk;
  for(trk = 0; trk <= 1; trk++)
    {
      char remote_url[status->url_size];
      const char*url = status->trk[trk].url;
      padico_string_t s_name = padico_string_new();
      padico_string_catf(s_name, "psp-minidriver-trk%d-tag%X", trk, status->slot->tag);
      const int name_len = padico_string_size(s_name);
      padico_addrdb_publish(node, padico_module_self_name(),
			    padico_string_get(s_name), name_len, url, status->url_size);
      padico_req_t req = padico_tm_req_new(NULL, NULL);
      padico_addrdb_get(node, padico_module_self_name(),
			padico_string_get(s_name), name_len, &remote_url[0], status->url_size, req);
      padico_tm_req_wait(req);
      struct puk_receptacle_NewMad_minidriver_s*r = &status->trk[trk].r;
      (*r->driver->connect)(r->_status, remote_url, status->url_size);
      padico_print("PSP-minidriver: trk#%d connection established \n", trk);
    }
  psp_minidriver_refill(status);
  psp_minidriver_conns_vect_push_back(&padico_minidriver.conns, status);

  if(!padico_minidriver.polling_thread_launched)
    {
      padico_rttask_t t = padico_rttask_new();
      t->kind = PADICO_RTTASK_RTTHREAD;
      t->background_fun = &psp_minidriver_worker;
      t->background_arg = NULL;
      padico_tm_rttask_submit(t);
      padico_minidriver.polling_thread_launched = 1;
    }
}

static padico_simplepackets_connection_t psp_minidriver_new_message(void*_status, padico_na_node_t node, void**_sbuf)
{
  struct psp_minidriver_status_s*const status = (struct psp_minidriver_status_s*)_status;
  psp_minidriver_lock(status);
  uint32_t*p_tag = (void*)status->sbuf;
  *p_tag = status->slot->tag;
  *_sbuf = p_tag + 1;
  status->header_sent = 0;
  return _status;
}

static void psp_minidriver_pack(void*_status, padico_simplepackets_connection_t _conn, const char*bytes, size_t size)
{
  struct psp_minidriver_status_s*const status = (struct psp_minidriver_status_s*)_status;
  const int trk = (size > status->threshold) ? 1 : 0;
  struct iovec v;
  if(!status->header_sent)
    {
      psp_minidriver_sendheader(status);
    }
  v.iov_base = (void*)bytes;
  v.iov_len = size;
  (*status->trk[trk].r.driver->send_post)(status->trk[trk].r._status, &v, 1);
  int rc = -1;
  do
    {
      rc = (*status->trk[trk].r.driver->send_poll)(status->trk[trk].r._status);
    }
  while(rc != 0);
}

static void psp_minidriver_end_message(void*_status, padico_simplepackets_connection_t _conn)
{
  struct psp_minidriver_status_s*const status = (struct psp_minidriver_status_s*)_status;
  if(!status->header_sent)
    {
      psp_minidriver_sendheader(status);
    }
  psp_minidriver_unlock(status);
}
#endif /* NMAD */
