/** @file
 * @brief Padico Simple Packets API
 * @ingroup PSP
 */

#ifndef PADICO_NA_PSP_H
#define PADICO_NA_PSP_H

#include <Padico/Puk.h>
#include <Padico/Topology.h>
#include <Padico/PadicoTM.h>

/** @defgroup NetAccessPSP API: PSP- Padico Simple Packets.
 * A simple low-level packet-based messaging API, for networks with
 * implicit connection.
 * @ingroup NetAccess
 */


/* ********************************************************* */
/* *** Interface: PadicoSimplePackets */

/** @addtogroup NetAccessPSP
 * @{
 */


/* ** Values and intervals for PSP channels */


/* implicit channels */
#define PADICO_PSP_CONTROL       ((uint32_t)0x01000000)  /**< Control/PSP */
#define PADICO_PSP_CIRCUIT       ((uint32_t)0x02000000)  /**< Circuit/PSP */
#define PADICO_PSP_VOPW          ((uint32_t)0x03000000)  /**< VLink-Packet */
#define PADICO_PSP_NEWMADICO     ((uint32_t)0x04000000)  /**< Newmadico */
/* reserved values for system use */
#define PADICO_PSP_RESERVED      ((uint32_t)0xF0000000)
#define PADICO_PSP_CLOSE_CHANNEL ((uint32_t)0xF1000000)  /**< [ reserved for internal use ] */
#define PADICO_PSP_SYSTEM        ((uint32_t)0xF2000000)  /**< [ reserved for internal use ] */

#define PADICO_PSP_ADAPTER_MASK  ((uint32_t)0xFF000000)
#define PADICO_PSP_INSTANCE_MASK ((uint32_t)0x00FFFFFF)

/** build a PSP tag from an adapter tag and an instance tag
 */
static inline uint32_t padico_psp_tag_build(uint32_t adapter_tag, uint32_t instance_tag)
{
  assert(!(adapter_tag & PADICO_PSP_INSTANCE_MASK));
  assert(!(instance_tag & PADICO_PSP_ADAPTER_MASK));
  return (adapter_tag | instance_tag);
}
/** get the 'adapter' (static) part from a PSP tag
 */
static inline uint32_t padico_psp_tag_get_adapter(uint32_t tag)
{
  return (tag & PADICO_PSP_ADAPTER_MASK);
}
/** get the 'instance' (dynamic) part from a PSP tag
 */
static inline uint32_t padico_psp_tag_get_instance(uint32_t tag)
{
  return (tag & PADICO_PSP_INSTANCE_MASK);
}

/** the function to pump data in PSP callbacks
 */
typedef void (*padico_simplepackets_pump_t)(void*token, void*bytes, size_t size);

/** A callback for PSP
 * @note the callback must call '*pump' with the same number and 
 * size of packets as the sender called 'pack'.
 */
typedef void (*padico_simplepackets_handler_t)(const void*header,
					       padico_na_node_t from,
					       void*key,
					       padico_simplepackets_pump_t pump,
					       void*token);

/** An outgoing connection object, supposed to be opaque for the user.
 */
typedef void*padico_simplepackets_connection_t;

/** driver descriptor for the PSP interface.
 */
struct padico_simplepackets_driver_s
{
  /** @brief init a new PSP instance
   * @param _instance the PSP instance status (driver-dependant)
   * @param tag       the tag identifying the PSP channel; statically allocated
   * @param label     the name of the channel; used for debugging and dynamically allocated PSPs
   * @param h_size    size of the header on the given channel; inout parameter: in=requested size; out=real size (out>=in)
   * @param handler   the function that will be called upon message reception
   * @param key       the parameter that will be given to the callback function
   */
  void(*init)(void*_instance, uint32_t tag, const char*label, size_t*h_size, padico_simplepackets_handler_t handler, void*key);

  /** @brief connects a remote node through PSP
   * @param _instance the PSP instance status
   * @param node      the remote node to connect to
   * @note PSP implementations _must_ support multiple calls to this
   *       function with the same 'node' parameter.
   */
  void(*connect)(void*_instance, padico_na_node_t node);

  /** @brief begin a new outgoing message
   * @param _instance  the PSP instance status
   * @param node       the recipient of the to-be-sent message
   * @param sbuf       the sending buffer for headers ('out' parameters)
   * @return           an opaque driver-dependant connection
   */
  padico_simplepackets_connection_t (*new_message)(void*_instance, padico_na_node_t node, void**sbuf);

  /** @brief add a new segment of data to an outgoing message
   * @param _instance  the PSP instance status
   * @param conn       the connection object of the currently built message
   * @param bytes      pointer to the data to send
   * @param size       size of the data to send
   */
  void(*pack)(void*_instance, padico_simplepackets_connection_t conn, const char*bytes, size_t size);

  /** @brief terminates an outgoing message
   * @param _instance  the PSP instance status
   * @param conn       the connection object of the message
   */
  void(*end_message)(void*_instance, padico_simplepackets_connection_t conn);
};


/** Attach the iface name to its driver */
PUK_IFACE_TYPE(PadicoSimplePackets, struct padico_simplepackets_driver_s);



/* *********************************************************
 * *** PSP directory ***************************************
 *
 * This directory is used for PSP demultiplexing of incoming
 * messages. This is a toolbox to be used by adapters 
 * exhibiting the PSP interface.
 */

#ifndef __cplusplus

struct padico_psp_slot_s
{
  padico_simplepackets_handler_t handler;
  void*key;
  uint32_t tag;
  size_t h_size;
  int ref_count;
};
typedef struct padico_psp_slot_s*padico_psp_slot_t;

struct padico_psp_directory_s
{
  pmarcel_spinlock_t lock;
  puk_hashtable_t table;
  padico_psp_slot_t cache;
};

static uint32_t padico_psp_tag_hash(const void*key)
{
  const uint32_t a = (uint32_t)((long)key);
  return a;
}
static int padico_psp_tag_eq(const void*key1, const void*key2)
{
  const uint32_t a = (uint32_t)((long)key1);
  const uint32_t b = (uint32_t)((long)key2);
  return a == b;
}


static inline void padico_psp_directory_init(struct padico_psp_directory_s*d)
{
  pmarcel_spin_init(&d->lock, 0);
  d->table = puk_hashtable_new(&padico_psp_tag_hash, &padico_psp_tag_eq);
  d->cache = NULL;
}

static inline padico_psp_slot_t padico_psp_slot_insert(struct padico_psp_directory_s*d,
						       padico_simplepackets_handler_t handler, void*key,
						       uint32_t tag, size_t h_size)
{
  pmarcel_spin_lock(&d->lock);
  padico_psp_slot_t slot = puk_hashtable_lookup(d->table, puk_long2ptr(tag));
  if(!slot)
    {
      slot = padico_malloc(sizeof(struct padico_psp_slot_s));
      *slot = (struct padico_psp_slot_s){ .handler = handler, .key = key, .tag = tag, .h_size = h_size, .ref_count = 1 };
      puk_hashtable_insert(d->table, puk_long2ptr(tag), slot);
    }
  else
    {
      assert(slot->handler == handler && slot->key == key);
      slot->ref_count++;
    }
  pmarcel_spin_unlock(&d->lock);
  return slot;
}

static inline void padico_psp_slot_remove(struct padico_psp_directory_s*d,
					  padico_psp_slot_t slot)
{
  pmarcel_spin_lock(&d->lock);
  slot->ref_count--;
  if(slot->ref_count <= 0)
    puk_hashtable_remove(d->table, puk_long2ptr(slot->tag));
  pmarcel_spin_unlock(&d->lock);
}

static inline padico_psp_slot_t padico_psp_slot_lookup(struct padico_psp_directory_s*d, uint32_t tag)
{
  pmarcel_spin_lock(&d->lock);
  padico_psp_slot_t slot = (d->cache && (d->cache->tag == tag)) ? d->cache : NULL;
  while(!slot)
    {
      slot = puk_hashtable_lookup(d->table, puk_long2ptr(tag));
      if(!slot)
	{
	  pmarcel_spin_unlock(&d->lock);
	  sched_yield();
	  pmarcel_spin_lock(&d->lock);
	}
    }
  d->cache = slot;
  pmarcel_spin_unlock(&d->lock);
  return slot;
}

/** @} */

#endif /* __cplusplus */

#endif /* PADICO_NA_PSP_H */
