/** @file
 * @brief BZ2 Packet Filter adapter
 */

#include <Padico/Module.h>
#include "PacketFilter.h"
#include <bzlib.h>

PADICO_MODULE_DECLARE(PacketFilter_BZ2);
PADICO_MODULE_INIT(bz2pf_init);

/* ********************************************************* */
/* *** BZ2 Packet Filter *********************************** */
/* ********************************************************* */


/** @defgroup PacketFilter_BZ2 adapter: PacketFilter-BZ2- a packet filter that performs bzip2 compression
 * @ingroup PadicoAdapter
 */

static void* bz2pf_instanciate(puk_instance_t assembly, puk_context_t context);
static void bz2pf_destroy(void*state);
static void bz2pf_encode(void*, const void*, size_t, void**, size_t*);
static void bz2pf_decode(void*, const void*, size_t, void*, size_t*);

/** instanciation facet for PacketFilter-BZ2
 * @ingroup PacketFilter_BZ2
 */
const static struct puk_adapter_driver_s bz2_adapter_driver =
  {
    .instanciate = &bz2pf_instanciate,
    .destroy     = &bz2pf_destroy
  };
/** 'PacketFilter' facet for PacketFilter-BZ2
 * @ingroup PacketFilter_BZ2
 */
const static struct packetfilter_driver_s bz2pf_driver = 
  {
    .pf_encode = &bz2pf_encode,
    .pf_decode = &bz2pf_decode
  };

/** instance status for PacketFilter-BZ2
 * @ingroup PacketFilter_BZ2
 */
struct bz2_state_s
{
  int block_size;  /* default: 5 */
  int work_factor; /* default: 0 (30) */
  int small;
};


static void* bz2pf_instanciate(puk_instance_t assembly, puk_context_t context)
{
  struct bz2_state_s*state = padico_malloc(sizeof(struct bz2_state_s));
  const char*block_size = puk_context_getattr(context, "block_size");
  const char*work_factor = puk_context_getattr(context, "work_factor");
  assert(block_size && work_factor);
  state->block_size = atoi(block_size);
  state->work_factor = atoi(work_factor);
  state->small = 0;
  return state;
}

static void bz2pf_destroy(void*state)
{
  assert(state != NULL);
  padico_free(state);
}


/* 'out' allocated by encode function
 */
static void bz2pf_encode(void*w,
		       const void*in, size_t in_size,
		       void**out, size_t* out_size)
{
  struct bz2_state_s*state = w;
  unsigned int s = in_size + in_size / 16 + 256;
  void*p = padico_malloc(s);
  int rc = BZ2_bzBuffToBuffCompress(p, &s, (void*)in, in_size,
				    state->block_size, 0, state->work_factor);
  if(rc != BZ_OK)
    {
      /* compression failed: don't compress */
      padico_fatal("BZ2: compression failed- rc=%d\n", rc);
      memcpy(p, in, in_size);
      s = in_size;
    }
  *out = p;
  *out_size = s;
}

/* 'out' allocated by caller
*/
static void bz2pf_decode(void*w,
			  const void*in, size_t in_size,
			  void*out, size_t* out_size)
{
  struct bz2_state_s*state = w;
  unsigned int s = *out_size;
  int rc = BZ2_bzBuffToBuffDecompress(out, &s, (void*)in, in_size,
				      state->small, 0);
  if(rc != BZ_OK)
    {
      padico_fatal("BZ2: CRC error while uncompressing- in_size=%d; out_size=%d; rc=%d\n",
		   (int)in_size, (int)*out_size, rc);
    }
  *out_size = s;
}

static int bz2pf_init(void)
{
  padico_print("PacketFilter-BZ2: using bzlib version %s\n", BZ2_bzlibVersion());

  puk_component_declare("PacketFilter-BZ2",
			puk_component_provides("PadicoAdapter", "adapter", &bz2_adapter_driver),
			puk_component_provides("PacketFilter", "pf", &bz2pf_driver),
			puk_component_attr("block_size", "5"),
			puk_component_attr("work_factor", "30"));
  return 0;
}
