/*
 * NewMadeleine
 * Copyright (C) 2012 (see AUTHORS file)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include <stdint.h>
#include <sys/uio.h>
#include <assert.h>

#include <nm_private.h>

#include <Padico/Module.h>
#include <Padico/Topology.h>
#include "shm.h"

#undef PROFILE

static struct
{
  int aggreg;
  int total;
}  nm_strat_shm_stats = { .total = 0, .aggreg = 0 };

/* Components structures:
 */

static int strat_shm_todo(void*_status, struct nm_gate*p_gate);
static int strat_shm_pack(void*_status, struct nm_pack_s*p_pack);
static int strat_shm_pack_ctrl(void*, struct nm_gate *, const union nm_so_generic_ctrl_header*);
static int strat_shm_try_and_commit(void*, struct nm_gate*);
static int strat_shm_rdv_accept(void*, struct nm_gate*, nm_len_t, int*, struct nm_rdv_chunk*);
static void strat_shm_proto(void*_status, struct nm_gate*p_gate, struct nm_pkt_wrap*p_pw, const void*ptr, nm_len_t len);

static const struct nm_strategy_iface_s nm_strat_shm_driver =
  {
    .pack            = &strat_shm_pack,
    .pack_ctrl       = &strat_shm_pack_ctrl,
    .try_and_commit  = &strat_shm_try_and_commit,
    .rdv_accept      = &strat_shm_rdv_accept,
    .flush           = NULL,
    .todo            = &strat_shm_todo,
    .proto           = &strat_shm_proto
};

static void*strat_shm_instanciate(puk_instance_t, puk_context_t);
static void strat_shm_destroy(void*);

static const struct puk_adapter_driver_s nm_strat_shm_adapter_driver =
  {
    .instanciate = &strat_shm_instanciate,
    .destroy     = &strat_shm_destroy
  };


/** Per-gate status for strat_shm instances
 */
struct nm_strat_shm 
{
  int nm_strat_shm_max_small;
};

PADICO_MODULE_COMPONENT(NewMad_Strategy_shm,
  puk_component_declare("NewMad_Strategy_shm",
			puk_component_provides("PadicoAdapter", "adapter", &nm_strat_shm_adapter_driver),
			puk_component_provides("NewMad_Strategy", "strat", &nm_strat_shm_driver),
			puk_component_attr("nm_strat_shm_max_small", "16342")));

/** node descriptor for strat_shm 
 */
struct nm_strat_shm_node_s
{
  padico_na_node_t node;             /**< the node described in this entry */
  struct nm_gate*p_gate;             /**< nm gate connected to the node */
  struct padico_shm_node_s*shm_node; /**< node in the shm segment- NULL if node is remote */
  int rank;                          /**< global nm rank */
  struct nm_strat_shm_node_s*leader; /**< leader node on the host (first node in the host, in nm launcher rank) */
};

static struct
{
  struct padico_shm_s*shm;  /**< shm segment for strat_shm */
  puk_hashtable_t gate_map; /**< nodes hashed by gate */
  puk_hashtable_t node_map; /**< nodes hash by uuid */
  struct nm_strat_shm_node_s*self_node;
} nm_strat_shm_global =
  {
    .shm = NULL,
    .gate_map = NULL,
    .node_map = NULL,
    .self_node = NULL
  };


struct strat_shm_header_s
{
  struct nm_strat_header strat_header;
  char from[PADICO_TOPO_UUID_SIZE];
  char dest[PADICO_TOPO_UUID_SIZE];
};

/* ********************************************************* */

#warning TODO- clean header
extern nm_gate_t newmadico_launcher_get_gate(int rank);
extern padico_na_node_t newmadico_launcher_get_node(int rank);

static void nm_strat_shm_init(void)
{
  if(nm_strat_shm_global.shm == NULL)
    {
      fprintf(stderr, "# nmad strat_shm: init- building node map.\n");
      nm_strat_shm_global.shm = padico_shm_init("nm_strat_shm");
      nm_strat_shm_global.node_map = puk_hashtable_new_string();
      nm_strat_shm_global.gate_map = puk_hashtable_new_ptr();
      int i;
      for(i = 0 ; ; i++)
	{
	  /* build gate <-> node mapping */
	  nm_gate_t p_gate = newmadico_launcher_get_gate(i);
	  padico_na_node_t node = newmadico_launcher_get_node(i);
	  if(p_gate == NULL || node == NULL)
	    break;
	  struct nm_strat_shm_node_s*leader = NULL;
	  puk_hashtable_enumerator_t e = puk_hashtable_enumerator_new(nm_strat_shm_global.node_map);
	  struct nm_strat_shm_node_s*n = puk_hashtable_enumerator_next_data(e);
	  while(n != NULL)
	    {
	      if(padico_na_node_gethost(node) == padico_na_node_gethost(n->node))
		{
		  leader = n->leader;
		  break;
		}
	      n = puk_hashtable_enumerator_next_data(e);
	    }
	  puk_hashtable_enumerator_delete(e);
	  struct nm_strat_shm_node_s*shm_node = padico_malloc(sizeof(struct nm_strat_shm_node_s));
	  if(leader == NULL)
	    {
	      leader = shm_node;
	    }
	  if(padico_na_node_gethost(node) == padico_na_node_gethost(padico_na_getlocalnode()))
	    {
	      shm_node->shm_node = padico_shm_directory_node_lookup(nm_strat_shm_global.shm, node);
	    }
	  else
	    {
	      shm_node->shm_node = NULL;
	    }
	  shm_node->node   = node;
	  shm_node->p_gate = p_gate;
	  shm_node->rank   = i;
	  shm_node->leader = leader;
	  puk_hashtable_insert(nm_strat_shm_global.gate_map, shm_node->p_gate, shm_node);
	  puk_hashtable_insert(nm_strat_shm_global.node_map, padico_na_node_getuuid(shm_node->node), shm_node);
	  fprintf(stderr, "# nmad strat_shm: #%d- %s -> gate = %p; route = %s\n", 
		  i, (const char*)padico_na_node_getuuid(node), p_gate, (const char*)padico_na_node_getuuid(leader->node));
	  if(node == padico_na_getlocalnode())
	    {
	      nm_strat_shm_global.self_node = shm_node;
	    }
	  /* build routing table */
	  if(node != padico_na_getlocalnode())
	    {
	      padico_na_network_vect_t networks = NULL;
	      networks = padico_na_network_vect_new();
	      padico_na_node_getnetworks(node, networks);
	      
	    }
	}
      assert(nm_strat_shm_global.self_node != NULL);
      fprintf(stderr, "# nmad strat_shm: init done; self = %s; leader = %d\n",
	      (const char*)padico_na_node_getuuid(nm_strat_shm_global.self_node->node),
	      nm_strat_shm_global.self_node == nm_strat_shm_global.self_node->leader);
    }
}

/* ********************************************************* */

/** Initialize the gate storage for strat_shm.
 */
static void*strat_shm_instanciate(puk_instance_t ai, puk_context_t context)
{
  struct nm_strat_shm *status = TBX_MALLOC(sizeof(struct nm_strat_shm));

  const char*nm_strat_shm_max_small = puk_instance_getattr(ai, "nm_strat_shm_max_small");
  status->nm_strat_shm_max_small = atoi(nm_strat_shm_max_small);
  NM_LOGF("[nm_strat_shm_max_small=%i]", status->nm_strat_shm_max_small);

  return (void*)status;
}

/** Cleanup the gate storage for strat_shm.
 */
static void strat_shm_destroy(void*status)
{
  TBX_FREE(status);
}



static void nm_strat_shm_pw_complete(struct nm_pkt_wrap*p_pw, struct nm_pw_completion_s*p_completion)
{
  const intptr_t _key = (intptr_t)p_completion->data.key;
  const int block = _key; 
  padico_shm_block_free(nm_strat_shm_global.shm, block);
  NM_TRACEF("block = %d; completed.", block);
}


/** Add a new control "header" to the flow of outgoing packets.
 *
 *  @param _status the strat_shm instance status.
 *  @param p_ctrl a pointer to the ctrl header.
 *  @return The NM status.
 */
static int strat_shm_pack_ctrl(void*_status, struct nm_gate*p_gate, nm_generic_header_t p_ctrl)
{
  struct nm_pkt_wrap*p_pw = NULL;
  nm_so_pw_alloc(NM_PW_GLOBAL_HEADER, &p_pw);
  nm_so_pw_add_control(p_pw, p_ctrl);
  nm_drv_t p_drv = nm_drv_default(p_gate);
  nm_core_post_send(p_gate, p_pw, NM_TRK_SMALL, p_drv);
  return NM_ESUCCESS;
}

static void nm_strat_shm_poll_direct(void)
{
  /* deliver direct packets (driver-bypass) */
  const int block_num = padico_shm_short_recv_poll(nm_strat_shm_global.shm);
  if(block_num > -1)
    {
      NM_TRACEF("direct send- received block %d\n", block_num);
      struct padico_shm_short_header_s*shm_header = padico_shm_block_get_ptr(nm_strat_shm_global.shm, block_num);
      const int from = shm_header->from;
      const int size = shm_header->size;
      const void*uuid = nm_strat_shm_global.shm->seg->directory.nodes[from].node_uuid;
      struct nm_strat_shm_node_s*shm_node = puk_hashtable_lookup(nm_strat_shm_global.node_map, uuid);
      assert(shm_node != NULL);
      struct nm_gate*p_from_gate = shm_node->p_gate;
      assert(p_from_gate != NULL);
      void*base = padico_shm_short_get_ptr(nm_strat_shm_global.shm, block_num);
      NM_TRACEF("direct decode- size = %d.\n", size);
      struct nm_pkt_wrap *p_pw = NULL;
      nm_so_pw_alloc(NM_PW_NOHEADER, &p_pw);
      nm_so_pw_add_raw(p_pw, base, size, 0 /* offset */);
      p_pw->header_ref_count++;
      nm_decode_header_chunk(p_from_gate->p_core, base, p_pw, p_from_gate);
      struct nm_pw_completion_s completion =
	{
	  .notifier = &nm_strat_shm_pw_complete,
	  .data.key = (void*)(intptr_t)block_num
	};
      nm_pw_add_completion(p_pw, &completion);
      nm_so_pw_dec_header_ref_count(p_pw);
      NM_TRACEF("decode done.\n");
    }
}

static int strat_shm_todo(void* _status, struct nm_gate*p_gate)
{ 
  if(nm_strat_shm_global.shm == NULL)
    nm_strat_shm_init();
  nm_strat_shm_poll_direct();
  if(nm_strat_shm_global.self_node == nm_strat_shm_global.self_node->leader)
    {
      struct nm_strat_shm_node_s*shm_node = puk_hashtable_lookup(nm_strat_shm_global.gate_map, p_gate);
      return padico_shm_block_lfqueue_empty(&nm_strat_shm_global.shm->seg->buffer.pending_blocks[shm_node->rank]);
    }
  return 0;
}


/** Handle a new packet submitted by the user code.
 *
 *  @note The strategy may already apply some optimizations at this point.
 *  @param p_gate a pointer to the gate object.
 *  @param tag the message tag.
 *  @param seq the fragment sequence number.
 *  @param data the data fragment pointer.
 *  @param len the data fragment length.
 *  @return The NM status.
 */
static int strat_shm_pack(void*_status, struct nm_pack_s*p_pack)
{
  struct nm_strat_shm*status = _status;
  if(nm_strat_shm_global.shm == NULL)
    nm_strat_shm_init();
  NM_LOG_IN();
  if(p_pack->status & NM_PACK_TYPE_CONTIGUOUS)
    {
      const void*data = p_pack->data;
      const int len = p_pack->len;
      if(len <= status->nm_strat_shm_max_small)
	{
#if 0
	  /* ******** */
	  const nm_len_t offset = 0;
	  struct nm_pkt_wrap *p_pw = NULL;
	  nm_gate_t p_gate = p_pack->p_gate;
	  nm_drv_t p_drv = nm_drv_default(p_gate);
	  nm_so_pw_alloc(NM_PW_GLOBAL_HEADER, &p_pw);
	  nm_so_pw_add_data(p_pw, p_pack, data, len, offset, NM_PW_GLOBAL_HEADER | NM_SO_DATA_USE_COPY);
	  nm_core_post_send(p_gate, p_pw, NM_TRK_SMALL, p_drv);
	  return NM_ESUCCESS;
	  /* ******** */
#endif

	  struct nm_strat_shm_node_s*shm_node = puk_hashtable_lookup(nm_strat_shm_global.gate_map, p_pack->p_gate);
	  assert(shm_node != NULL);
	  assert(len + NM_SO_DATA_HEADER_SIZE <= PADICO_SHM_BLOCKSIZE);
	  /* ** small pack */
	  const int block = padico_shm_block_get(nm_strat_shm_global.shm);
	  if(shm_node->leader == nm_strat_shm_global.self_node->leader)
	    {
	      /* ** direct send through shm- no external driver involved ' */
	      NM_TRACEF("direct send.\n");
	      void*ptr = padico_shm_short_get_ptr(nm_strat_shm_global.shm, block);
	      struct nm_so_data_header*data_header = ptr;
	      data_header->proto_id = NM_PROTO_DATA | NM_PROTO_LAST; /* XXX TODO- short data */
	      data_header->tag_id   = p_pack->tag;
	      data_header->seq      = p_pack->seq;
	      data_header->flags    = NM_PROTO_FLAG_LASTCHUNK | NM_PROTO_FLAG_ALIGNED;
	      data_header->len      = p_pack->len;
	      data_header->chunk_offset = 0;
	      data_header->skip     = 0;
	      /* copy data */
	      memcpy(ptr + NM_SO_DATA_HEADER_SIZE, p_pack->data, p_pack->len);
	      struct padico_shm_node_s*dest = shm_node->shm_node;
	      assert(dest != NULL);
	      padico_shm_short_send_commit(nm_strat_shm_global.shm, dest, block, NM_SO_DATA_HEADER_SIZE + nm_so_aligned(p_pack->len));
	      NM_TRACEF("direct send- done.\n");
	    }
	  else
	    {
	      /* ** indirect send through shm */
	      const char*dest_uuid = (const char*)padico_na_node_getuuid(shm_node->node);
	      assert(dest_uuid != NULL);
	      void*ptr = padico_shm_block_get_ptr(nm_strat_shm_global.shm, block);
	      /* pack strat header */
	      struct strat_shm_header_s*shm_header = ptr;
	      shm_header->strat_header.proto_id = NM_PROTO_STRAT;
	      shm_header->strat_header.size = sizeof(struct strat_shm_header_s) + NM_SO_DATA_HEADER_SIZE + p_pack->len;
	      memcpy(shm_header->dest, dest_uuid, PADICO_TOPO_UUID_SIZE);
	      memcpy(shm_header->from, padico_na_node_getuuid(padico_na_getlocalnode()), PADICO_TOPO_UUID_SIZE);
	      /* pack data header */
	      struct nm_so_data_header*data_header = ptr + sizeof(struct strat_shm_header_s);
	      data_header->proto_id = NM_PROTO_DATA | NM_PROTO_LAST;
	      data_header->tag_id   = p_pack->tag;
	      data_header->seq      = p_pack->seq;
	      data_header->flags    = NM_PROTO_FLAG_LASTCHUNK | NM_PROTO_FLAG_ALIGNED; /* XXX */
	      data_header->len      = p_pack->len;
	      data_header->chunk_offset = 0;
	      data_header->skip     = 0;
	      /* copy data */
	      memcpy(ptr + sizeof(struct strat_shm_header_s) + NM_SO_DATA_HEADER_SIZE, p_pack->data, p_pack->len);
	      __sync_synchronize();
	      int rc = padico_shm_block_lfqueue_enqueue(&nm_strat_shm_global.shm->seg->buffer.pending_blocks[shm_node->leader->rank], block);
	      assert(rc == 0);
	    }
	  /* immediate completion */
	  const struct nm_core_event_s event =
	    {
	      .status = NM_STATUS_PACK_COMPLETED,
	      .p_pack = p_pack
	    };
	  nm_core_status_event(p_pack->p_gate->p_core, &event, &p_pack->status);
	}
      else
	{
	  nm_tactic_pack_rdv(p_pack, data, len, 0);
	}
    }
  else
    {
      return -NM_ENOTIMPL;
    }
  NM_LOG_OUT();
  return NM_ESUCCESS;
}

/** Compute and apply the best possible packet rearrangement, then
 *  return next packet to send.
 *
 *  @param p_gate a pointer to the gate object.
 *  @return The NM status.
 */
static int strat_shm_try_and_commit(void*_status, struct nm_gate *p_gate)
{
#ifdef PROFILE
  tbx_tick_t t1, t2;
  int posted = 0, active = 0;

  TBX_GET_TICK(t1);
#endif

  if(nm_strat_shm_global.shm == NULL)
    nm_strat_shm_init();
  NM_LOG_IN();
  nm_strat_shm_poll_direct();
  if(nm_strat_shm_global.self_node == nm_strat_shm_global.self_node->leader)
    {
      struct nm_strat_shm_node_s*shm_node = puk_hashtable_lookup(nm_strat_shm_global.gate_map, p_gate);
      assert(shm_node != NULL);
      struct padico_shm_block_lfqueue_s*queue = &nm_strat_shm_global.shm->seg->buffer.pending_blocks[shm_node->rank];
      if(shm_node->leader == shm_node)
	{
	  /* src and dest are node leaders */
	  nm_drv_t p_drv = nm_drv_default(p_gate);
	  struct nm_gate_drv*p_gdrv = nm_gate_drv_get(p_gate, p_drv);
	  NM_TRACEF("node is leader. trk #0 state: %d", p_gdrv->active_send[NM_TRK_SMALL]);
#ifdef PROFILE
	  active = p_gdrv->active_send[NM_TRK_SMALL];
#endif
	  if(p_gdrv->active_send[NM_TRK_SMALL] == 0)
	    {
	      struct nm_pkt_wrap *p_pw = NULL;
	      int block = padico_shm_block_lfqueue_dequeue_single_reader(queue);
	      while(block != -1)
		{
		  NM_TRACEF("processing block = %d", block);
		  nm_strat_shm_stats.total++;
		  struct strat_shm_header_s*header = padico_shm_block_get_ptr(nm_strat_shm_global.shm, block);
		  if(p_pw == NULL)
		    nm_so_pw_alloc(NM_PW_NOHEADER, &p_pw);
		  nm_so_pw_add_raw(p_pw, header, header->strat_header.size, 0 /* chunk_offset */);
		  struct nm_pw_completion_s completion =
		    {
		      .notifier = &nm_strat_shm_pw_complete,
		      .data.key = (void*)(intptr_t)block
		    };
		  nm_pw_add_completion(p_pw, &completion);
		  block = padico_shm_block_lfqueue_dequeue_single_reader(queue);
		  if(block != -1)
		    {
		      struct strat_shm_header_s*next_header = padico_shm_block_get_ptr(nm_strat_shm_global.shm, block);
		      if(next_header->strat_header.size >= nm_so_pw_remaining_data(p_pw))
			{
			  padico_shm_block_lfqueue_enqueue(queue, block);
			  header->strat_header.proto_id |= NM_PROTO_LAST;
			  break;
			}
		      else
			{
			  nm_strat_shm_stats.aggreg++;
			  /*
			  if(nm_strat_shm_stats.aggreg % 1000 == 0)
			    fprintf(stderr, "# nm_strat_shm: aggreg= %d / %d (%2.2g%%)\n",
				    nm_strat_shm_stats.aggreg, nm_strat_shm_stats.total,
				    (100.0*nm_strat_shm_stats.aggreg/nm_strat_shm_stats.total));
			  */
			}
		    }
		  else
		    {
		      header->strat_header.proto_id |= NM_PROTO_LAST;
		      break;
		    }
		}
	      if(p_pw != NULL)
		{
		  nm_core_post_send(p_gate, p_pw, NM_TRK_SMALL, p_drv);
#ifdef PROFILE
		  posted = 1;
#endif
		}
	    }
	}
    }
  NM_LOG_OUT();

#ifdef PROFILE
  TBX_GET_TICK(t2);
#define MULT 10017
  static int toto = 0;
  static const int mult = MULT;
  static tbx_tick_t orig;
  static double tab[MULT];
  const double t = TBX_TIMING_DELAY(t1, t2);
  tab[toto % mult] = t;
  if(toto++ % mult == 0)
    {
      const double period = TBX_TIMING_DELAY(orig, t2) / mult;
      fprintf(stderr, "try_and_commit: %6.1f nsec.; period = %6.1f; p_gate = %16p; posted = %d; active = %d\n", t*1000.0,  period * 1000.0, p_gate, posted, active);
      orig = t2;

      int i;
      double min = tab[0];
      double max = tab[0];
      double cumul = 0.0;
      for(i = 0; i < mult; i++)
	{
	  cumul += tab[i];
	  if(tab[i] > max)
	    max = tab[i];
	  if(tab[i] < min)
	    min = tab[i];
	}
      const double avg = cumul / mult;
      fprintf(stderr, "          min = %6.1f nsec; max = %6.1f nsec.; avg = %6.1f nsec.\n", min * 1000.0, max * 1000.0, avg * 1000.0);

    }
#endif
  return NM_ESUCCESS;
}

static void strat_shm_proto(void*_status, struct nm_gate*p_gate, struct nm_pkt_wrap*p_pw, const void*ptr, nm_len_t len)
{
  NM_LOG_IN();
  const struct strat_shm_header_s*shm_header = ptr;
  if(memcmp(shm_header->dest, padico_na_node_getuuid(padico_na_getlocalnode()), PADICO_TOPO_UUID_SIZE) == 0)
    {
      char from_uuid[PADICO_TOPO_UUID_SIZE + 1];
      memcpy(from_uuid, shm_header->from, PADICO_TOPO_UUID_SIZE);
      from_uuid[PADICO_TOPO_UUID_SIZE] = '\0';
      struct nm_strat_shm_node_s*shm_from_node = puk_hashtable_lookup(nm_strat_shm_global.node_map, from_uuid);
      assert(shm_from_node != NULL);
      struct nm_gate*p_from_gate = shm_from_node->p_gate;
      assert(p_from_gate != NULL);
      nm_decode_header_chunk(p_gate->p_core, ptr + sizeof(struct strat_shm_header_s), p_pw, p_from_gate);
    }
  else
    {
      NM_TRACEF("forwarding message.\n");
      char dest_uuid[PADICO_TOPO_UUID_SIZE + 1];
      memcpy(dest_uuid, shm_header->dest, PADICO_TOPO_UUID_SIZE);
      dest_uuid[PADICO_TOPO_UUID_SIZE] = '\0';
      struct nm_strat_shm_node_s*shm_dest_node = puk_hashtable_lookup(nm_strat_shm_global.node_map, dest_uuid);
      const int block = padico_shm_block_get(nm_strat_shm_global.shm);
      void*block_ptr = padico_shm_short_get_ptr(nm_strat_shm_global.shm, block);
      memcpy(block_ptr, ptr, len);
      struct padico_shm_node_s*dest = shm_dest_node->shm_node;
      assert(dest != NULL);
      padico_shm_short_send_commit(nm_strat_shm_global.shm, dest, block, len);
    }
  NM_LOG_OUT();
}

/** Accept or refuse a RDV on the suggested (driver/track/gate).
 *
 *  @warning @p drv_id and @p trk_id are IN/OUT parameters. They initially
 *  hold values "suggested" by the caller.
 *  @param p_gate a pointer to the gate object.
 *  @param drv_id the suggested driver id.
 *  @param trk_id the suggested track id.
 *  @return The NM status.
 */
static int strat_shm_rdv_accept(void*_status, struct nm_gate *p_gate, nm_len_t len,
				int*nb_chunks, struct nm_rdv_chunk*chunks)
{
  *nb_chunks = 1;
  nm_drv_t p_drv = nm_drv_default(p_gate);
  struct nm_gate_drv*p_gdrv = nm_gate_drv_get(p_gate, p_drv);
  if(p_gdrv->active_recv[NM_TRK_LARGE] == 0)
    {
      /* The large-packet track is available! */
      chunks[0].len = len;
      chunks[0].p_drv = p_drv;
      chunks[0].trk_id = NM_TRK_LARGE;
      return NM_ESUCCESS;
    }
  else
    /* postpone the acknowledgement. */
    return -NM_EAGAIN;
}
