/* @file Timing.h
 * @brief Timing service
 * @note includes a lot of copy/paste from PM2 tbx_tick
 * @author Alexandre Denis and Raymond Namyst
 */

#ifndef PADICO_TIMING_H
#define PADICO_TIMING_H


#include <sys/types.h>
#include <sys/time.h>

#include <Padico/Puk.h>

#ifdef __cplusplus
extern "C" {
#endif

/* Most compilers don't set __i586__ as they should do.
 * We assert that __i386__ ==> __i586__
 * How to fix?- maybe it could autoconf'ed
 */
#if defined(__i386__) || defined(__pentium__) || defined(__pentiumpro__) || defined(__i586__) || defined(__i686__) || defined(__k6__) || defined(__k7__) || defined(__x86_64__)
#define USE_PENTIUM_TIMER
#else
#undef USE_PENTIUM_TIMER
#endif

typedef union 
{
  uint64_t tick;
  struct 
  {
    uint32_t low, high;
  } sub;
  struct timeval timev;
} padico_timing_t;

#define PADICO_TIMING_GET_TICK(t)       padico_timing_get_tick(&(t))
#define PADICO_TIMING_TICK_DIFF(t1, t2) padico_timing_diff_tick(&(t1), &(t2))

/** capture the current timer */
extern void      padico_timing_get_tick (padico_timing_t*t);

/** computes a time interval- result in ticks (=cpu cycle with Pentium timing, =microseconds else) */
extern int64_t padico_timing_diff_tick(padico_timing_t*t1, padico_timing_t*t2);

/** computes a time interval in microseconds */
extern double    padico_timing_diff_usec(padico_timing_t*t1, padico_timing_t*t2);

/** convert a time interval from ticks into microseconds */
extern double    padico_timing_tick2usec(int64_t t);


  /* *** Profiling ***************************************** */

  struct padico_timing_profiler_s
  {
    int num_points;         /**< number of profiling points per iteration */
    int num_iter;           /**< number of iterations beeing profiled */
    int current_iter;       /**< current iteration number */
    padico_timing_t*points; /**< profiling points data (size = num_iter x num_points array) */
    const char*filename;    /**< filename where to flush profiling data */
    const char**labels;     /**< labels of points (size = num_points) */
  };
  typedef struct padico_timing_profiler_s* padico_timing_profiler_t;

  extern padico_timing_profiler_t padico_timing_profiler_new(int num_points,
							     int num_iter,
							     int ignore,
							     const char*filename);

  extern void padico_timing_profile_point_label(padico_timing_profiler_t p, int n_point, const char*label);

#define padico_timing_profile_point(P, N) padico_timing_profile_point_label((P), (N), __FUNCTION__)

  extern void padico_timing_profile_start(padico_timing_profiler_t p);

  extern void padico_timing_profile_flush(padico_timing_profiler_t p);

#undef PADICO_TIMING_DO_FLUSH

#ifdef __cplusplus
} /* extern "C" */
#endif


#endif /* PADICO_TIMING_H */

