/** @file
 * @brief Puk main functions
 * @ingroup Puk
 * @note Puk is not a true PadicoTM module. It is declared as
 * a module for compatibility with padico_module_self()
 */

#include <stdlib.h>
#include <setjmp.h>

#include "Module.h"
#include "Puk-internals.h"

PADICO_MODULE_HOOK(Puk); /* only a hook here. 
			  * For some obscure reasons, the real
			  * declaration is in puk-mod.c
			  */

/******************************************************/

static struct
{
  void (*lock_acquire)(void);
  void (*lock_release)(void);
  int lock_level;
  char  hostname[256];
  const char*root;
  int initialized, finalized;
} puk = { .lock_acquire = NULL, .lock_release = NULL, .lock_level = 0, 
	  .root = NULL,
	  .initialized = 0,
	  .finalized = 0
        };

/******************************************************/

void puk_lock_set_handlers(void(*_acquire)(void), void(*_release)(void))
{
  puk.lock_acquire = _acquire;
  puk.lock_release = _release;
}

void __puk_lock_acquire(const char*func)
{
  /* lazy init */
  if(!puk.initialized)
    padico_puk_init(0, NULL);
  padico_out(80, "%s locking...\n", func);
  if(puk.lock_acquire)
    (*puk.lock_acquire)();
  assert(puk.lock_level >= 0);
  puk.lock_level++;
  padico_out(80, "%s locked - level=%d\n", func, puk.lock_level);
}
void __puk_lock_release(const char*func)
{
  padico_out(80, "%s unlocking- level=%d\n", func, puk.lock_level - 1);
  assert(puk.lock_level > 0);
  puk.lock_level--;
  if(puk.lock_release)
    (*puk.lock_release)();
}

int padico_puk_initialized(void)
{
  return puk.initialized;
}

void padico_puk_init(int argc, char**argv)
{
  if(puk.initialized)
    return; /* Puk already initialized */
  if(puk.finalized)
    padico_fatal("Puk: trying to initialize Puk after shutdown.\n");
  if(padico_getenv("PADICO_VERBOSE"))
    {
      /* welcome banner */
      fprintf(stderr, "** Welcome to " PADICO_BUILD_STRING "\n");
    }
  if(gethostname(puk.hostname, 256))
    padico_fatal("Puk: cannot get gethostname()\n");
  const char *_padico_root = padico_getenv("PADICO_ROOT");
  puk.root = _padico_root?:PADICO_ROOT;
  puk.initialized = 1;

  /* low-level init */
  padico_puk_xml_init();        /* XML parser */
  padico_puk_components_init(); /* component framework */
  padico_puk_builtin_init();    /* driver: 'builtin' */
  padico_puk_mod_init();        /* module management */
  padico_puk_pkg_init();        /* driver: 'pkg' */
  padico_puk_debug_init();      /* debugging tools */
  padico_puk_lib_init();        /* routines for dynamic libraries */
  padico_puk_monitor_init();    /* monitoring functions ('lsmod' and the likes) */
  padico_puk_binary_init();     /* driver: 'binary' */
}

void padico_puk_start(void)
{
  padico_mod_t startup_mod = NULL;
  padico_rc_t rc = NULL;
  /* load the startup module */
  const char* startup_name = "PadicoTM";
  padico_print("Puk: loading startup module '%s'.\n", startup_name);
  rc = padico_puk_mod_resolve(&startup_mod, startup_name);
  if(padico_rc_iserror(rc))
    {
      padico_rc_show(rc);
      padico_fatal("Puk: startup module '%s' not found.\n", startup_name);
    }
  rc = padico_puk_mod_load(startup_mod);
  if(padico_rc_iserror(rc))
    {
      padico_rc_show(rc);
      padico_fatal("Puk: error while loading startup module.\n");
    }
  /* The startup mod is supposed to take the control of the process.
   * We never reach here.
   */
  padico_fatal("Puk: invalid startup module. It should never return from init!\n");
}

void padico_puk_shutdown(void)
{
  if(puk.initialized && !puk.finalized)
    {
      padico_print("Puk: shutdown.\n");
      padico_puk_mod_finalize();
      padico_puk_components_finalize();
      padico_puk_debug_finalize();
      padico_puk_xml_finalize();
      puk.root = NULL;
      puk.initialized = 0;
      puk.finalized = 1;
    }
}

void puk_notimplemented()
{
  padico_fatal("Function not implemented\n");
}

/******************************************************/


void padico_setenv(const char *var, const char *val)
{
#ifdef HAVE_SETENV
  setenv(var, val, 1);
#else
  int l;
  char *p;
  l = strlen(var)+strlen(val)+2;
  p = (char*) padico_malloc(l);
  strcpy(p, var);
  strcat(p, "=");
  strcat(p, val);
  putenv(p);
#endif
}

const char*padico_getenv(const char*label)
{
  const char*value = getenv(label);
  return value;
}

void padico_cleanenv()
{
  const char*padico_root = padico_getenv("PADICO_ROOT");
  padico_string_t new_lib = padico_string_new();
  padico_string_t padico_lib = padico_string_new();
  padico_string_catf(padico_lib, "%s/lib", padico_root);
  char*saveptr;
  char*lib = getenv("LD_LIBRARY_PATH");
  char*token = strtok_r(lib, ":", &saveptr);
  while(token != NULL)
    {
      if(strstr(token, padico_string_get(padico_lib)) == NULL)
	{
	  padico_string_catf(new_lib, "%s:", token);
	}
      token = strtok_r(NULL, ":", &saveptr);
    }
  padico_setenv("LD_LIBRARY_PATH", padico_string_get(new_lib));
  padico_string_delete(new_lib);
  padico_string_delete(padico_lib);
}

const char*padico_hostname(void)
{
  return puk.hostname;
}

const char*padico_root(void)
{
  return puk.root;
}


